import glob
from os.path import dirname, basename, isfile

# Loads all .py file in the current dir making them available for top level import
modules = glob.glob(dirname(__file__) + "/*.py")
__all__ = [
    basename(f)[:-3] for f in modules if isfile(f) and not f.endswith("__init__.py")
]



# -*- coding: utf-8 -*-

import abc
import logging
from typing import Iterable, Union, List, Callable, Dict, Any, Tuple
from datetime import datetime

import pandas as pd
import numpy as np


logger = logging.getLogger(__name__)


class InsufficientDataError(ValueError):
    pass


class GordoBaseDataset:

    _params: Dict[Any, Any] = dict()  # provided by @capture_args on child's __init__
    _metadata: Dict[Any, Any] = dict()

    @abc.abstractmethod
    def get_data(
        self,
    ) -> Tuple[Union[np.ndarray, pd.DataFrame], Union[np.ndarray, pd.DataFrame]]:
        """
        Return X, y data as numpy or pandas' dataframes given current state
        """

    @abc.abstractmethod
    def to_dict(self) -> dict:
        """
        Serialize this object into a dict representation, which can be used to
        initialize a new object using :func:`~GordoBaseDataset.from_dict`

        Returns
        -------
        dict
                """
        if not hasattr(self, "_params"):
            raise AttributeError(
                f"Failed to lookup init parameters, ensure the "
                f"object's __init__ is decorated with 'capture_args'"
            )
        # Update dict with the class
        params = self._params
        params["type"] = self.__class__.__name__
        for key, value in params.items():
            if hasattr(value, "to_dict"):
                params[key] = value.to_dict()
        return params

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, config: Dict[str, Any]) -> "GordoBaseDataset":
        """
        Construct the dataset using a config from :func:`~GordoBaseDataset.to_dict`
        """
        from gordo.machine.dataset import datasets

        Dataset = getattr(datasets, config.get("type", "TimeSeriesDataset"))
        if Dataset is None:
            raise TypeError(f"No dataset of type '{config['type']}'")

        # TODO: Here for compatibility, but @compate should take care of it, remove later
        if "tags" in config:
            config["tag_list"] = config.pop("tags")
        config.setdefault("target_tag_list", config["tag_list"])
        return Dataset(**config)

    @abc.abstractmethod
    def get_metadata(self):
        """
        Get metadata about the current state of the dataset
        """

    def join_timeseries(
        self,
        series_iterable: Iterable[pd.Series],
        resampling_startpoint: datetime,
        resampling_endpoint: datetime,
        resolution: str,
        aggregation_methods: Union[str, List[str], Callable] = "mean",
    ) -> pd.DataFrame:
        """

        Parameters
        ----------
        series_iterable: Iterable[pd.Series]
            An iterator supplying series with time index
        resampling_startpoint: datetime.datetime
            The starting point for resampling. Most data frames will not have this
            in their datetime index, and it will be inserted with a NaN as the value.
            The resulting NaNs will be removed, so the only important requirement for this is
            that this resampling_startpoint datetime must be before or equal to the first
            (earliest) datetime in the data to be resampled.
        resampling_endpoint: datetime.datetime
            The end point for resampling. This datetime must be equal to or after the last datetime in the
            data to be resampled.
        resolution: str
            The bucket size for grouping all incoming time data (e.g. "10T")
            Available strings come from https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
        aggregation_methods: Union[str, List[str], Callable]
            Aggregation method(s) to use for the resampled buckets. If a single
            resample method is provided then the resulting dataframe will have names
            identical to the names of the series it got in. If several
            aggregation-methods are provided then the resulting dataframe will
            have a multi-level column index, with the series-name as the first level,
            and the aggregation method as the second level.
            See :py:func::`pandas.core.resample.Resampler#aggregate` for more
            information on possible aggregation methods.

        Returns
        -------
        pd.DataFrame
            A dataframe without NaNs, a common time index, and one column per
            element in the dataframe_generator. If multiple aggregation methods
            are provided then the resulting dataframe will have a multi-level column
            index with series-names as top-level and aggregation-method as second-level.

        """
        resampled_series = []
        missing_data_series = []

        key = "tag_loading_metadata"
        self._metadata[key] = dict()

        for series in series_iterable:
            self._metadata[key][series.name] = dict(original_length=len(series))
            try:
                resampled = GordoBaseDataset._resample(
                    series,
                    resampling_startpoint=resampling_startpoint,
                    resampling_endpoint=resampling_endpoint,
                    resolution=resolution,
                    aggregation_methods=aggregation_methods,
                )
            except IndexError:
                missing_data_series.append(series.name)
            else:
                resampled_series.append(resampled)
                self._metadata[key][series.name].update(
                    dict(resampled_length=len(resampled))
                )
        if missing_data_series:
            raise InsufficientDataError(
                f"The following features are missing data: {missing_data_series}"
            )

        joined_df = pd.concat(resampled_series, axis=1, join="inner")

        # Before returning, delete all rows with NaN, they were introduced by the
        # insertion of NaNs in the beginning of all timeseries
        dropped_na = joined_df.dropna()

        self._metadata[key]["aggregate_metadata"] = dict(
            joined_length=len(joined_df), dropped_na_length=len(dropped_na)
        )
        return dropped_na

    @staticmethod
    def _resample(
        series: pd.Series,
        resampling_startpoint: datetime,
        resampling_endpoint: datetime,
        resolution: str,
        aggregation_methods: Union[str, List[str], Callable] = "mean",
    ):
        """
        Takes a single series and resamples it.
        See :class:`gordo.machine.dataset.base.GordoBaseDataset.join_timeseries`
        """

        startpoint_sametz = resampling_startpoint.astimezone(tz=series.index[0].tzinfo)
        endpoint_sametz = resampling_endpoint.astimezone(tz=series.index[0].tzinfo)

        if series.index[0] > startpoint_sametz:
            # Insert a NaN at the startpoint, to make sure that all resampled
            # indexes are the same. This approach will "pad" most frames with
            # NaNs, that will be removed at the end.
            startpoint = pd.Series(
                [np.NaN], index=[startpoint_sametz], name=series.name
            )
            series = startpoint.append(series)
            logging.debug(
                f"Appending NaN to {series.name} " f"at time {startpoint_sametz}"
            )

        elif series.index[0] < resampling_startpoint:
            msg = (
                f"Error - for {series.name}, first timestamp "
                f"{series.index[0]} is before the resampling start point "
                f"{startpoint_sametz}"
            )
            logging.error(msg)
            raise RuntimeError(msg)

        if series.index[-1] < endpoint_sametz:
            endpoint = pd.Series([np.NaN], index=[endpoint_sametz], name=series.name)
            series = series.append(endpoint)
            logging.debug(
                f"Appending NaN to {series.name} " f"at time {endpoint_sametz}"
            )
        elif series.index[-1] > endpoint_sametz:
            msg = (
                f"Error - for {series.name}, last timestamp "
                f"{series.index[-1]} is later than the resampling end point "
                f"{endpoint_sametz}"
            )
            logging.error(msg)
            raise RuntimeError(msg)

        logging.debug("Head (3) and tail(3) of dataframe to be resampled:")
        logging.debug(series.head(3))
        logging.debug(series.tail(3))

        resampled = series.resample(resolution, label="left").agg(aggregation_methods)
        # If several aggregation methods are provided, agg returns a dataframe
        # instead of a series. In this dataframe the column names are the
        # aggregation methods, like "max" and "mean", so we have to make a
        # multi-index with the series-name as the top-level and the
        # aggregation-method as the lower-level index.
        # For backwards-compatibility we *dont* return a multi-level index
        # when we have a single resampling method.
        if isinstance(resampled, pd.DataFrame):  # Several aggregation methods provided
            resampled.columns = pd.MultiIndex.from_product(
                [[series.name], resampled.columns], names=["tag", "aggregation_method"]
            )
        return resampled.fillna(method="ffill")



import os

from unittest.mock import patch
import pytest

import dateutil.parser

from gordo.machine.dataset.data_provider.ncs_reader import NcsReader, NcsParquetLookup
from gordo.machine.dataset.data_provider.providers import DataLakeProvider
from gordo.machine.dataset.sensor_tag import normalize_sensor_tags
from gordo.machine.dataset.sensor_tag import SensorTag


class AzureDLFileSystemMock:
    def info(self, file_path):
        info = {"length": os.path.getsize(file_path)}
        if os.path.isfile(file_path):
            info["type"] = "FILE"
        elif os.path.isdir(file_path):
            info["type"] = "DIRECTORY"
        return info

    def open(self, file_path, mode):
        return open(file_path, mode)


@pytest.fixture
def ncs_reader():
    return NcsReader(AzureDLFileSystemMock())


@pytest.fixture
def dates():
    return (
        dateutil.parser.isoparse("2000-01-01T08:56:00+00:00"),
        dateutil.parser.isoparse("2001-09-01T10:01:00+00:00"),
    )


@pytest.mark.parametrize(
    "tag_to_check",
    [normalize_sensor_tags(["TRC-123"])[0], SensorTag("XYZ-123", "1776-TROC")],
)
def test_can_handle_tag_ok(tag_to_check, ncs_reader):
    assert ncs_reader.can_handle_tag(tag_to_check)


@pytest.mark.parametrize(
    "tag_to_check", [SensorTag("TRC-123", None), SensorTag("XYZ-123", "123-XXX")]
)
def test_can_handle_tag_notok(tag_to_check, ncs_reader):
    assert not ncs_reader.can_handle_tag(tag_to_check)


def test_can_handle_tag_unknow_prefix_raise(ncs_reader):
    with pytest.raises(ValueError):
        ncs_reader.can_handle_tag(normalize_sensor_tags(["XYZ-123"])[0])


def test_can_handle_tag_non_supported_asset_with_base_path(ncs_reader):
    tag = SensorTag("WEIRD-123", "UNKNOWN-ASSET")
    assert not ncs_reader.can_handle_tag(tag)

    ncs_reader_with_base = NcsReader(
        AzureDLFileSystemMock(), dl_base_path="/this/is/a/base/path"
    )
    assert ncs_reader_with_base.can_handle_tag(tag)


def test_load_series_need_base_path(ncs_reader, dates):
    tag = SensorTag("WEIRD-123", "BASE-PATH-ASSET")
    with pytest.raises(ValueError):
        for _ in ncs_reader.load_series(dates[0], dates[1], [tag]):
            pass

    path_to_weird_base_path_asset = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "data",
        "datalake",
        "base_path_asset",
    )
    ncs_reader_with_base = NcsReader(
        AzureDLFileSystemMock(), dl_base_path=path_to_weird_base_path_asset
    )
    for tag_series in ncs_reader_with_base.load_series(dates[0], dates[1], [tag]):
        assert len(tag_series) == 20


@pytest.mark.parametrize(
    "bad_tag_lists", [["TRC-123"], [{"name": "TRC-123", "asset": None}]]
)
def test_load_series_tag_as_string_fails(bad_tag_lists, dates, ncs_reader):
    with pytest.raises(AttributeError):
        for _ in ncs_reader.load_series(dates[0], dates[1], bad_tag_lists):
            pass


def test_load_series_need_asset_hint(dates, ncs_reader):
    with pytest.raises(ValueError):
        for _ in ncs_reader.load_series(
            dates[0], dates[1], [SensorTag("XYZ-123", None)]
        ):
            pass

    path_to_xyz = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "data", "datalake", "gordoplatform"
    )
    with patch(
        "gordo.machine.dataset.data_provider.ncs_reader.NcsReader.ASSET_TO_PATH",
        {"gordoplatform": path_to_xyz},
    ):
        valid_tag_list_with_asset = [SensorTag("XYZ-123", "gordoplatform")]
        for frame in ncs_reader.load_series(
            dates[0], dates[1], valid_tag_list_with_asset
        ):
            assert len(frame) == 20


@patch(
    "gordo.machine.dataset.data_provider.ncs_reader.NcsReader.ASSET_TO_PATH",
    {
        "1776-troc": os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "data", "datalake"
        )
    },
)
def test_load_series_known_prefix(dates, ncs_reader):
    valid_tag_list_no_asset = normalize_sensor_tags(["TRC-123", "TRC-321"])
    for frame in ncs_reader.load_series(dates[0], dates[1], valid_tag_list_no_asset):
        assert len(frame) == 20


@patch(
    "gordo.machine.dataset.data_provider.ncs_reader.NcsReader.ASSET_TO_PATH",
    {
        "1776-troc": os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "data", "datalake"
        )
    },
)
@pytest.mark.parametrize(
    "start_date, end_date, frame_len",
    [
        [
            dateutil.parser.isoparse("2000-01-01T08:56:00+00:00"),
            dateutil.parser.isoparse("2021-09-01T10:01:00+00:00"),
            20,
        ],
        [
            # Test something not valid (where there could not be any data)
            dateutil.parser.isoparse("2025-01-01T00:00:00+00:00"),
            dateutil.parser.isoparse("2030-01-01T00:00:00+00:00"),
            0,
        ],
    ],
)
def test_load_series_invalid_year(start_date, end_date, frame_len, ncs_reader):
    valid_tag_list = normalize_sensor_tags(["TRC-123"])
    frame = next(ncs_reader.load_series(start_date, end_date, valid_tag_list))
    assert len(frame) == frame_len


def test_ncs_reader_valid_tag_path():
    with pytest.raises(FileNotFoundError):
        NcsReader._verify_tag_path_exist(AzureDLFileSystemMock(), "not/valid/path")


@patch(
    "gordo.machine.dataset.data_provider.ncs_reader.NcsReader.ASSET_TO_PATH",
    {
        "1776-troc": os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "data", "datalake"
        )
    },
)
def test_load_series_dry_run(dates, ncs_reader):
    valid_tag_list_no_asset = normalize_sensor_tags(["TRC-123", "TRC-321"])
    for frame in ncs_reader.load_series(
        dates[0], dates[1], valid_tag_list_no_asset, dry_run=True
    ):
        assert len(frame) == 0


@patch(
    "gordo.machine.dataset.data_provider.ncs_reader.NcsReader.ASSET_TO_PATH",
    {
        "1776-troc": os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "data", "datalake"
        )
    },
)
@pytest.mark.parametrize("remove_status_codes", [[], [0]])
def test_load_series_with_filter_bad_data(dates, remove_status_codes):

    ncs_reader = NcsReader(
        AzureDLFileSystemMock(), remove_status_codes=remove_status_codes
    )

    valid_tag_list = normalize_sensor_tags(["TRC-322"])
    series_gen = ncs_reader.load_series(dates[0], dates[1], valid_tag_list)
    # Checks if the bad data from the files under tests/gordo/data_provider/data/datalake/TRC-322
    # are filtered out. 20 rows exists, 5 of then have the value 0.

    n_expected = 15 if remove_status_codes != [] else 20
    assert all(len(series) == n_expected for series in series_gen)


@pytest.mark.parametrize("remove_status_codes", [[], [0]])
def test_ncs_reader_kwargs_contains_remove_status_codes(remove_status_codes):
    # Creates a DataLakeProvider with remove_status_codes as kwargs
    data_provider = DataLakeProvider(
        interactive=False, remove_status_codes=remove_status_codes
    )

    # Set the data_provider's client to the AzureDLFileSystemMock as interactive can be False.
    data_provider.client = AzureDLFileSystemMock()
    # Get the ncs_reader from data_provider.
    ncs_reader = data_provider._get_sub_dataproviders()[0]

    # Cheks that the kwargs remove_status_codes has been passed to the sub_provider
    expected = [] if remove_status_codes == [] else [0]
    assert ncs_reader.remove_status_codes == expected


@patch(
    "gordo.machine.dataset.data_provider.ncs_reader.NcsReader.ASSET_TO_PATH",
    {
        "1776-troc": os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "data", "datalake"
        )
    },
)
def test_parquet_files_lookup(dates):
    ncs_reader = NcsReader(AzureDLFileSystemMock(), remove_status_codes=[0])

    valid_tag_list = normalize_sensor_tags(["TRC-323"])
    series_gen = ncs_reader.load_series(dates[0], dates[1], valid_tag_list)
    tags_series = [v for v in series_gen]
    assert len(tags_series) == 1
    trc_323_series = tags_series[0]
    assert trc_323_series.name == "TRC-323"
    assert trc_323_series.dtype.name == "float64"
    assert len(trc_323_series) == 20


def test_get_file_lookups():
    with pytest.raises(ValueError):
        NcsReader.get_file_lookups([])
    with pytest.raises(ValueError):
        NcsReader.get_file_lookups(["excel"])
    file_lookups = NcsReader.get_file_lookups(["parquet"])
    assert len(file_lookups) == 1
    assert isinstance(file_lookups[0], NcsParquetLookup)


@patch(
    "gordo.machine.dataset.data_provider.ncs_reader.NcsReader.ASSET_TO_PATH",
    {
        "1776-troc": os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "data", "datalake"
        )
    },
)
def test_with_conflicted_file_types(dates):
    ncs_reader = NcsReader(AzureDLFileSystemMock(), remove_status_codes=[0])

    valid_tag_list = normalize_sensor_tags(["TRC-324"])
    series_gen = ncs_reader.load_series(dates[0], dates[1], valid_tag_list)
    tags_series = [v for v in series_gen]
    assert len(tags_series) == 1
    trc_324_series = tags_series[0]
    # Parquet file should be with 15 rows
    assert len(trc_324_series) == 15


@patch(
    "gordo.machine.dataset.data_provider.ncs_reader.NcsReader.ASSET_TO_PATH",
    {
        "1776-troc": os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "data", "datalake"
        )
    },
)
def test_with_conflicted_file_types_with_preferable_csv(dates):
    ncs_reader = NcsReader(
        AzureDLFileSystemMock(), remove_status_codes=[0], lookup_for=["csv"]
    )

    valid_tag_list = normalize_sensor_tags(["TRC-324"])
    series_gen = ncs_reader.load_series(dates[0], dates[1], valid_tag_list)
    tags_series = [v for v in series_gen]
    assert len(tags_series) == 1
    trc_324_series = tags_series[0]
    # CSV file should be with 1 row
    assert len(trc_324_series) == 1



import pandas as pd
import numpy as np

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import BinaryIO, Optional


@dataclass
class TimeSeriesColumns:
    """
    Names of columns witch is used in time series datasets
    """

    datetime_column: str
    value_column: str
    status_column: Optional[str] = None

    @property
    def columns(self):
        columns = [self.datetime_column, self.value_column]
        if self.status_column is not None:
            columns.append(self.status_column)
        return columns


class FileType(metaclass=ABCMeta):
    """
    :class:`pandas.DataFrame` reader from the different file types
    """

    file_extension: Optional[str] = None

    @abstractmethod
    def read_df(self, f: BinaryIO) -> pd.DataFrame:
        """
        Read `DataFrame` from file object

        Parameters
        ----------
        f : BinaryIO
            File object
        """
        raise NotImplementedError()


class CsvFileType(FileType):

    file_extension: Optional[str] = ".csv"

    def __init__(
        self, header: list, time_series_columns: TimeSeriesColumns, sep: str = ";"
    ):
        """
        Create `DataFrame` reader for CSV files

        Parameters
        ----------
        header: list
            List of all columns in CSV file
        time_series_columns: TimeSeriesColumns
        sep: str
            Delimiter for columns in CSV file
        """
        self.header = header
        self.time_series_columns = time_series_columns
        self.sep = sep

    def read_df(self, f: BinaryIO) -> pd.DataFrame:
        datetime_column = self.time_series_columns.datetime_column
        value_column = self.time_series_columns.value_column
        return pd.read_csv(
            f,
            sep=self.sep,
            header=None,
            names=self.header,
            usecols=self.time_series_columns.columns,
            dtype={value_column: np.float32},
            parse_dates=[datetime_column],
            date_parser=lambda col: pd.to_datetime(col, utc=True),
            index_col=datetime_column,
        )


class ParquetFileType(FileType):

    file_extension: str = ".parquet"

    def __init__(self, time_series_columns: TimeSeriesColumns):
        """
        Create `DataFrame` reader for Parquet files

        Parameters
        ----------
        time_series_columns: TimeSeriesColumns
        """
        self.time_series_columns = time_series_columns

    def read_df(self, f: BinaryIO) -> pd.DataFrame:
        columns = self.time_series_columns.columns
        datetime_column = self.time_series_columns.datetime_column
        df = pd.read_parquet(f, engine="pyarrow", columns=columns).set_index(
            datetime_column
        )
        df.index = pd.to_datetime(df.index, utc=True)
        return df



import os
import time
import logging
from contextlib import contextmanager
from threading import Lock
from typing import List

import numpy as np
import pandas as pd

import requests
from influxdb import InfluxDBClient


from gordo.machine.model import models
from gordo.machine.dataset.sensor_tag import SensorTag

logger = logging.getLogger(__name__)


TEST_SERVER_MUTEXT = Lock()


def wait_for_influx(max_wait=120, influx_host="localhost:8086"):
    """Waits up to `max_wait` seconds for influx at `influx_host` to start up.

    Checks by pinging inflix at /ping.

    Parameters
    ----------
    influx_host : str
        Where is influx?
    max_wait : int
        How many seconds to wait in total for influx to start up

    Returns
    -------
    bool
        True if influx started up, False if it did not start during the `max_wait`
        period.

    """
    healtcheck_endpoint = f"http://{influx_host}/ping?verbose=true"
    before_time = time.perf_counter()
    influx_ok = False
    while not influx_ok and time.perf_counter() - before_time < max_wait:
        try:
            code = requests.get(
                healtcheck_endpoint, timeout=1, proxies={"https": "", "http": ""}
            ).status_code
            logger.debug(f"Influx gave code {code}")
            influx_ok = code == 200
        except requests.exceptions.ConnectionError:
            influx_ok = False
        time.sleep(0.5)
    if (time.perf_counter() - before_time) < max_wait:
        logger.info("Found that influx started")
        return True
    else:
        logger.warning("Found that influx never started")
        return False


@contextmanager
def temp_env_vars(**kwargs):
    """
    Temporarily set the process environment variables
    """
    _env = os.environ.copy()

    for key in kwargs:
        os.environ[key] = kwargs[key]

    yield

    os.environ.clear()
    os.environ.update(_env)


class InfluxDB:
    """
    Simple interface to a running influx.
    """

    def __init__(
        self,
        sensors: List[SensorTag],
        db_name: str,
        user: str,
        password: str,
        measurement: str,
    ):
        self.sensors = sensors
        self.db_name = db_name
        self.user = user
        self.password = password
        self.measurement = measurement

    def reset(self):
        """
        Set the db to contain the default data
        """
        # Seed database with some records
        influx_client = InfluxDBClient(
            "localhost",
            8086,
            self.user,
            self.password,
            self.db_name,
            proxies={"http": "", "https": ""},
        )

        # Drop and re-create the database
        influx_client.drop_database(self.db_name)
        influx_client.create_database(self.db_name)

        dates = pd.date_range(
            start="2016-01-01", periods=2880, freq="min"
        )  # Minute intervals for 2 days

        logger.info("Seeding database")
        for sensor in self.sensors:
            logger.info(f"Loading tag: {sensor.name}")
            points = np.random.random(size=dates.shape[0])
            data = [
                {
                    "measurement": self.measurement,
                    "tags": {"tag": sensor.name},
                    "time": f"{date}",
                    "fields": {"Value": point},
                }
                for point, date in zip(points, dates)
            ]
            influx_client.write_points(data)


def get_model(config):
    type = config.get("type", "")
    Model = getattr(models, type, None)
    if Model is None:
        raise ValueError(
            f'Type of model: "{type}" either not provided or not supported'
        )
    return Model(**config)



import logging
import os
import traceback
from typing import Tuple
import warnings

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "0.0.0"


def _parse_version(version: str) -> Tuple[int, ...]:
    """
    Takes a string which starts with standard major.minor.patch.
    and returns the split of major and minor version as integers

    Parameters
    ----------
    version: str
        The semantic version string

    Returns
    -------
    Tuple[int, int]
        major and minor versions
    """
    return tuple(int(i) for i in version.split(".")[:2])


MAJOR_VERSION, MINOR_VERSION = _parse_version(__version__)

try:
    # FIXME(https://github.com/abseil/abseil-py/issues/99)
    # FIXME(https://github.com/abseil/abseil-py/issues/102)
    # Unfortunately, many libraries that include absl (including Tensorflow)
    # will get bitten by double-logging due to absl's incorrect use of
    # the python logging library:
    #   2019-07-19 23:47:38,829 my_logger   779 : test
    #   I0719 23:47:38.829330 139904865122112 foo.py:63] test
    #   2019-07-19 23:47:38,829 my_logger   779 : test
    #   I0719 23:47:38.829469 139904865122112 foo.py:63] test
    # The code below fixes this double-logging.  FMI see:
    #   https://github.com/tensorflow/tensorflow/issues/26691#issuecomment-500369493

    import absl.logging

    logging.root.removeHandler(absl.logging._absl_handler)
    absl.logging._warn_preinit_stderr = False

except Exception:
    warnings.warn(f"Failed to fix absl logging bug {traceback.format_exc()}")
    pass



# -*- coding: utf-8 -*-

import abc

from datetime import datetime
from typing import Iterable, List, Optional

import pandas as pd

from gordo.machine.dataset.sensor_tag import SensorTag


class GordoBaseDataProvider(object):
    @abc.abstractmethod
    def load_series(
        self,
        train_start_date: datetime,
        train_end_date: datetime,
        tag_list: List[SensorTag],
        dry_run: Optional[bool] = False,
    ) -> Iterable[pd.Series]:
        """
        Load the required data as an iterable of series where each
        contains the values of the tag with time index

        Parameters
        ----------
        train_start_date: datetime
            Datetime object representing the start of fetching data
        train_end_date: datetime
            Datetime object representing the end of fetching data
        tag_list: List[SensorTag]
            List of tags to fetch, where each will end up being its own dataframe
        dry_run: Optional[bool]
            Set to true to perform a "dry run" of the loading.
            Up to the implementations to determine what that means.

        Returns
        -------
        Iterable[pd.Series]
        """
        ...

    @abc.abstractmethod
    def can_handle_tag(self, tag: SensorTag):
        """
        Returns true if the dataprovider thinks it can possibly read this tag.
        Typically checks if the asset part of the tag is known to the reader.

        Parameters
        ----------
        tag: SensorTag - Dictionary with a "tag" key and optional "asset"

        Returns
        -------
        bool

        """
        ...

    @abc.abstractmethod
    def to_dict(self):
        """
        Serialize this object into a dict representation, which can be used to
        initialize a new object after popping 'type' from the dict.

        Returns
        -------
        dict
        """
        if not hasattr(self, "_params"):
            raise AttributeError(
                f"Failed to lookup init parameters, ensure the "
                f"object's __init__ is decorated with 'capture_args'"
            )
        # Update dict with the class
        params = self._params
        params["type"] = self.__class__.__name__
        return params

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, config: dict) -> "GordoBaseDataProvider":
        from gordo.machine.dataset.data_provider import providers

        Provider = getattr(providers, config.get("type", "DataLakeProvider"))
        if Provider is None:
            raise TypeError(f"No data provider of type '{config['type']}'")
        return Provider(**config)



# -*- coding: utf-8 -*-
import logging
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
import timeit
from typing import Iterable, List, Optional
from urllib.parse import quote

import pandas as pd
from azure.datalake.store import core

from gordo.machine.dataset.data_provider.base import GordoBaseDataProvider
from gordo.machine.dataset.data_provider.file_type import (
    FileType,
    CsvFileType,
    ParquetFileType,
    TimeSeriesColumns,
)
from gordo.machine.dataset.data_provider.azure_utils import is_file
from gordo.machine.dataset.sensor_tag import SensorTag
from gordo.util import capture_args

logger = logging.getLogger(__name__)

time_series_columns = TimeSeriesColumns("Time", "Value", "Status")


class NcsFileLookup:
    def __init__(self, file_type: FileType):
        """
        Creates a files finder.

        Notes
        -----
        This implementation related only for :class:`gordo.machine.dataset.data_provider.ndc_reader.NcsReader` not for :class:`gordo.machine.dataset.data_provider.iroc_reader.IrocReader`

        Parameters
        ----------
        file_type : FileType
        """
        self.file_type = file_type

    def lookup(
        self, client: core.AzureDLFileSystem, dir_path: str, tag_name: str, year: int
    ) -> Optional[str]:
        """
        Find files with a given file type in Azure Data Lake

        Parameters
        ----------
        client: core.AzureDLFileSystem
            Azure Data Lake client
        dir_path: str
            Base directory for finding files
        tag_name
            File tag (encoded version)
        year
            File year
        """
        raise NotImplementedError()


class NcsCsvLookup(NcsFileLookup):
    """
    Finder for CSV files
    """

    def __init__(self):
        header = ["Sensor", "Value", "Time", "Status"]
        super().__init__(CsvFileType(header, time_series_columns))

    def lookup(
        self, client: core.AzureDLFileSystem, dir_path: str, tag_name: str, year: int
    ) -> Optional[str]:
        file_extension = self.file_type.file_extension
        path = f"{dir_path}/{tag_name}_{year}{file_extension}"
        return path if is_file(client, path) else None


class NcsParquetLookup(NcsFileLookup):
    """
    Finder for Parquet files
    """

    def __init__(self):
        super().__init__(ParquetFileType(time_series_columns))

    def lookup(
        self, client: core.AzureDLFileSystem, dir_path: str, tag_name: str, year: int
    ) -> Optional[str]:
        file_extension = self.file_type.file_extension
        path = f"{dir_path}/parquet/{tag_name}_{year}{file_extension}"
        return path if is_file(client, path) else None


class NcsReader(GordoBaseDataProvider):
    ASSET_TO_PATH = {
        # Paths on the datalake with problematic tag naming schemes (e.g. misplaced,
        # varying delimiters, non-unique, etc.) are commented with their assioted tag prefixes
        "1100-sfa": "/raw/corporate/IMS Statfjord/sensordata/1100-SFA",  # None
        "1101-sfb": "/raw/corporate/IMS Statfjord/sensordata/1101-SFB",  # None
        "1102-sfc": "/raw/corporate/IMS Statfjord/sensordata/1102-SFC",  # None
        "1110-gfa": "/raw/corporate/Aspen MS - IP21 Gullfaks A/sensordata/1110-GFA",
        "1111-gfb": "/raw/corporate/Aspen MS - IP21 Gullfaks B/sensordata/1111-GFB",
        "1112-gfc": "/raw/corporate/Aspen MS - IP21 Gullfaks C/sensordata/1112-GFC",
        "1120-vfr": "/raw/corporate/Aspen MS - IP21 Veslefrikk/sensordata/1120-VFR",
        "1125-kvb": "/raw/corporate/PI System Operation Norway/sensordata/1125-KVB",
        "1130-troa": "/raw/corporate/Aspen MS - IP21 Troll A/sensordata/1130-TROA",
        "1138-val": "/raw/corporate/PI System Operation Norway/sensordata/1138-VAL",
        "1140-sla": "/raw/corporate/PI System Manager Sleipner/sensordata/1140-SLA",  # None
        "1141-slt": "/raw/corporate/PI System Manager Sleipner/sensordata/1141-SLT",  # None
        "1142-slb": "/raw/corporate/PI System Manager Sleipner/sensordata/1142-SLB",  # None
        "1163-gdr": "/raw/corporate/PI System Manager Sleipner/sensordata/1163-GDR",  # None
        "1170-hd": "/raw/corporate/PI System Operation North/sensordata/1170-HD",
        "1175-kri": "/raw/corporate/PI System Operation North/sensordata/1175-KRI",  # kri.
        "1175-kris": "/raw/corporate/PI System Operation North/sensordata/1175-KRIS",  # kri.
        "1180-nor": "/raw/corporate/PI System Operation North/sensordata/1180-NOR",
        "1190-asga": "/raw/corporate/PI System Operation North/sensordata/1190-ASGA",
        "1191-asgb": "/raw/corporate/PI System Operation North/sensordata/1191-ASGB",  # asga. asgb-
        "1192-asgs": "/raw/corporate/PI System Operation North/sensordata/1192-ASGS",  # asgb.
        "1218-gkr": "/raw/corporate/PI System Manager Sleipner/sensordata/1218-GKR",  # None + 1218.
        "1219-aha": "/raw/corporate/PI System Operation Mam/sensordata/1219-AHA",
        "1220-sna": "/raw/corporate/IMS Snorre A/sensordata/1220-SNA",  # None
        "1221-snb": "/raw/corporate/IMS Snorre B/sensordata/1221-SNB",  # None
        "1230-vis": "/raw/corporate/Aspen MS - IP21 Visund/sensordata/1230-VIS",
        "1294-pera": "/raw/corporate/Aspen MS - IP21 Peregrino/sensordata/1294-PERA",  # per.
        "1295-pera": "/raw/corporate/Aspen MS - IP21 Peregrino/sensordata/1295-PERA",  # per.
        "1298-perb": "/raw/corporate/Aspen MS - IP21 Peregrino/sensordata/1298-PERB",  # per.
        "1299-per": "/raw/corporate/Aspen MS - IP21 Peregrino/sensordata/1299-PER",  # per.
        "1340-met": "/raw/corporate/PI System Operation Norway/sensordata/1340-MET",
        "1380-sno": "/raw/corporate/Aspen MS - IP21 Hammerfest/sensordata/1380-SNO",  # 25haxxx_
        "1755-gra": "/raw/corporate/Aspen MS - IP21 Grane/sensordata/1755-GRA",
        "1760-hea": "/raw/corporate/PI System Operation Norway/sensordata/1760-HEA",
        "1765-osc": "/raw/corporate/Aspen MS - IP21 Oseberg C/sensordata/1765-OSC",
        "1766-oss": "/raw/corporate/Aspen MS - IP21 Oseberg South/sensordata/1766-OSS",
        "1767-ose": "/raw/corporate/Aspen MS - IP21 Oseberg East/sensordata/1767-OSE",
        "1772-osa": "/raw/corporate/Aspen MS - IP21 Oseberg Field Center/sensordata/1772-OSA",  # osf.
        "1774-osd": "/raw/corporate/Aspen MS - IP21 Oseberg Field Center/sensordata/1774-OSD",  # osf.
        "1775-trob": "/raw/corporate/Aspen MS - IP21 Troll B/sensordata/1775-TROB",
        "1776-troc": "/raw/corporate/Aspen MS - IP21 Troll C/sensordata/1776-TROC",
        "1886-mara": "/raw/corporate/PI System Operation Mam/sensordata/1886-MARA",
        "1900-jsv": "/raw/corporate/PI System Operation Johan Sverdrup/sensordata/1900-JSV",
        "1901-jsv": "/raw/corporate/PI System Operation Johan Sverdrup/sensordata/1901-JSV",
        "1902-jsv": "/raw/corporate/PI System Operation Johan Sverdrup/sensordata/1902-JSV",
        "1903-jsv": "/raw/corporate/PI System Operation Johan Sverdrup/sensordata/1903-JSV",
        "1904-jsv": "/raw/corporate/PI System Operation Johan Sverdrup/sensordata/1904-JSV",
    }

    ALL_FILE_LOOKUPS = OrderedDict(
        (("parquet", NcsParquetLookup()), ("csv", NcsCsvLookup()))
    )

    @classmethod
    def get_file_lookups(cls, lookup_for: List[str]) -> List[NcsFileLookup]:
        if not lookup_for:
            raise ValueError("'lookup_for' must not be empty")
        file_lookups = []
        for lookup_name in lookup_for:
            try:
                file_lookups.append(cls.ALL_FILE_LOOKUPS[lookup_name])
            except KeyError:
                raise ValueError(
                    "Wrong lookup type '%s' in 'lookup_for' property", lookup_name
                )
        return file_lookups

    @capture_args
    def __init__(
        self,
        client: core.AzureDLFileSystem,
        threads: Optional[int] = 1,
        remove_status_codes: Optional[list] = [0],
        dl_base_path: Optional[str] = None,
        lookup_for: Optional[List[str]] = None,
        **kwargs,
    ):
        """
        Creates a reader for tags from the Norwegian Continental Shelf. Currently
        only supports a small subset of assets.

        Parameters
        ----------
        threads : Optional[int]
            Number of threads to use. If None then use 1 thread
        remove_status_codes: Optional[list]
            Removes data with Status code(s) in the list. By default it removes data
            with Status code 0.
        dl_base_path: Optional[str]
            Base bath used to override the asset to path dictionary. Useful for demos
            and other non-production settings.
        lookup_for:  Optional[List[str]]
            List of file finders by the file type name. Value by default: ``['parquet', 'csv']``

        Notes
        -----
        `lookup_for` provide list sorted by priority. It means that for value ``['csv', 'parquet']``
        the reader will prefer to find CSV files over Parquet

        """
        self.client = client
        self.threads = threads
        self.remove_status_codes = remove_status_codes
        self.dl_base_path = dl_base_path

        if lookup_for is None:
            file_lookups = list(self.ALL_FILE_LOOKUPS.values())
        else:
            file_lookups = self.get_file_lookups(lookup_for)
        self.file_lookups = file_lookups
        logger.info(f"Starting NCS reader with {self.threads} threads")

    def can_handle_tag(self, tag: SensorTag):
        """
        Implements GordoBaseDataProvider, see base class for documentation
        """
        return (
            self.dl_base_path is not None
            or NcsReader.base_path_from_asset(tag.asset) is not None
        )

    def load_series(
        self,
        train_start_date: datetime,
        train_end_date: datetime,
        tag_list: List[SensorTag],
        dry_run: Optional[bool] = False,
    ) -> Iterable[pd.Series]:
        """
        See GordoBaseDataProvider for documentation
        """
        if train_end_date < train_start_date:
            raise ValueError(
                f"NCS reader called with train_end_date: {train_end_date} before train_start_date: {train_start_date}"
            )
        adls_file_system_client = self.client

        years = range(train_start_date.year, train_end_date.year + 1)

        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            fetched_tags = executor.map(
                lambda tag: self.read_tag_files(
                    adls_file_system_client=adls_file_system_client,
                    tag=tag,
                    years=years,
                    dry_run=dry_run,
                    remove_status_codes=self.remove_status_codes,
                    dl_base_path=self.dl_base_path,
                ),
                tag_list,
            )

            for tag_frame_all_years in fetched_tags:
                filtered = tag_frame_all_years[
                    (tag_frame_all_years.index >= train_start_date)
                    & (tag_frame_all_years.index < train_end_date)
                ]
                yield filtered

    @staticmethod
    def _verify_tag_path_exist(
        adls_file_system_client: core.AzureDLFileSystem, path: str
    ):
        """
        Verify that the tag path exists, if not the `adls_file_system_client.info` will raise a FileNotFound error.

        Parameters
        ----------
        adls_file_system_client: core.AzureDLFileSystem
            the AzureDLFileSystem client to use
        path : str
            Path of tag to be checked if exists.
        """
        adls_file_system_client.info(f"{path}")

    def read_tag_files(
        self,
        adls_file_system_client: core.AzureDLFileSystem,
        tag: SensorTag,
        years: range,
        dry_run: Optional[bool] = False,
        remove_status_codes: Optional[list] = [0],
        dl_base_path: Optional[str] = None,
    ) -> pd.Series:
        """
        Download tag files for the given years into dataframes,
        and return as one dataframe.

        Parameters
        ----------
        adls_file_system_client: core.AzureDLFileSystem
            the AzureDLFileSystem client to use
        tag: SensorTag
            the tag to download data for
        years: range
            range object providing years to include
        dry_run: Optional[bool]
            if True, don't download data, just check info, log, and return
        remove_status_codes: Optional[list]
            Removes data with Status code(s) in the list. By default it removes data
            with Status code 0.
        dl_base_path: Optional[str]
            Base bath used to override the asset to path dictionary. Useful for demos
            and other non-production settings.

        Returns
        -------
        pd.Series:
            Series with all years for one tag.
        """
        tag_base_path = (
            dl_base_path if dl_base_path else NcsReader.base_path_from_asset(tag.asset)
        )

        if not tag_base_path:
            raise ValueError(f"Unable to find base path from tag {tag} ")
        all_years = []
        logger.info(f"Downloading tag: {tag} for years: {years}")
        tag_name_encoded = quote(tag.name, safe=" ")

        NcsReader._verify_tag_path_exist(
            adls_file_system_client, f"{tag_base_path}/{tag_name_encoded}/"
        )

        dir_path = f"{tag_base_path}/{tag_name_encoded}"
        for year in years:
            file_path = None
            file_lookup = None
            for v in self.file_lookups:
                file_path = v.lookup(
                    adls_file_system_client, dir_path, tag_name_encoded, year
                )
                if file_path is not None:
                    file_lookup = v
                    break
            if file_lookup is None:
                continue
            file_type = file_lookup.file_type
            logger.info(f"Parsing file {file_path}")

            try:
                info = adls_file_system_client.info(file_path)
                file_size = info.get("length") / (1024 ** 2)
                logger.debug(f"File size for file {file_path}: {file_size:.2f}MB")

                if dry_run:
                    logger.info("Dry run only, returning empty frame early")
                    return pd.Series()
                before_downloading = timeit.default_timer()
                with adls_file_system_client.open(file_path, "rb") as f:
                    df = file_type.read_df(f)
                    df = df.rename(columns={"Value": tag.name})
                    df = df[~df["Status"].isin(remove_status_codes)]
                    df.sort_index(inplace=True)
                    all_years.append(df)
                    logger.info(
                        f"Done in {(timeit.default_timer()-before_downloading):.2f} sec {file_path}"
                    )

            except FileNotFoundError as e:
                logger.debug(f"{file_path} not found, skipping it: {e}")

        try:
            combined = pd.concat(all_years)
        except Exception as e:
            logger.debug(f"Not able to concatinate all years: {e}.")
            return pd.Series(name=tag.name, data=None)

        # There often comes duplicated timestamps, keep the last
        if combined.index.duplicated().any():
            combined = combined[~combined.index.duplicated(keep="last")]

        return combined[tag.name]

    @staticmethod
    def base_path_from_asset(asset: str):
        """
        Resolves an asset code to the datalake basepath containing the data.
        Returns None if it does not match any of the asset codes we know.
        """
        if not asset:
            return None

        logger.debug(f"Looking for match for asset {asset}")
        asset = asset.lower()
        if asset not in NcsReader.ASSET_TO_PATH:
            logger.warning(
                f"Could not find match for asset {asset} in the list of "
                f"supported assets: {NcsReader.ASSET_TO_PATH.keys()}"
            )
            return None

        logger.debug(
            f"Found asset code {asset}, returning {NcsReader.ASSET_TO_PATH[asset]}"
        )
        return NcsReader.ASSET_TO_PATH[asset]



# -*- coding: utf-8 -*-
import logging

from typing import Tuple, List, Dict, Union, Optional, Iterable, Callable, Sequence
from datetime import datetime
from dateutil.parser import isoparse
from functools import wraps

import pandas as pd
import numpy as np

from gordo.machine.dataset.data_provider.providers import (
    RandomDataProvider,
    DataLakeProvider,
)
from gordo.machine.dataset.base import GordoBaseDataset, InsufficientDataError
from gordo.machine.dataset.data_provider.base import GordoBaseDataProvider
from gordo.machine.dataset.filter_rows import pandas_filter_rows
from gordo.machine.dataset.sensor_tag import SensorTag
from gordo.machine.dataset.sensor_tag import normalize_sensor_tags
from gordo.util import capture_args
from gordo.machine.validators import (
    ValidTagList,
    ValidDatetime,
    ValidDatasetKwargs,
    ValidDataProvider,
)

logger = logging.getLogger(__name__)


class InsufficientDataAfterRowFilteringError(InsufficientDataError):
    pass


def compat(init):
    """
    __init__ decorator for compatibility where the Gordo config file's ``dataset`` keys have
    drifted from what kwargs are actually expected in the given dataset. For example,
    using `train_start_date` is common in the configs, but :class:`~TimeSeriesDataset`
    takes this parameter as ``train_start_date``, as well as :class:`~RandomDataset`

    Renames old/other acceptable kwargs to the ones that the dataset type expects
    """

    @wraps(init)
    def wrapper(*args, **kwargs):
        renamings = {
            "from_ts": "train_start_date",
            "to_ts": "train_end_date",
            "tags": "tag_list",
        }
        for old, new in renamings.items():
            if old in kwargs:
                kwargs[new] = kwargs.pop(old)
        return init(*args, **kwargs)

    return wrapper


class TimeSeriesDataset(GordoBaseDataset):

    train_start_date = ValidDatetime()
    train_end_date = ValidDatetime()
    tag_list = ValidTagList()
    target_tag_list = ValidTagList()
    data_provider = ValidDataProvider()
    kwargs = ValidDatasetKwargs()

    @compat
    @capture_args
    def __init__(
        self,
        train_start_date: Union[datetime, str],
        train_end_date: Union[datetime, str],
        tag_list: Sequence[Union[str, Dict, SensorTag]],
        target_tag_list: Optional[Sequence[Union[str, Dict, SensorTag]]] = None,
        data_provider: Union[GordoBaseDataProvider, dict] = DataLakeProvider(),
        resolution: Optional[str] = "10T",
        row_filter: str = "",
        aggregation_methods: Union[str, List[str], Callable] = "mean",
        row_filter_buffer_size: int = 0,
        asset: Optional[str] = None,
        default_asset: Optional[str] = None,
        n_samples_threshold: int = 0,
        **_kwargs,
    ):
        """
        Creates a TimeSeriesDataset backed by a provided dataprovider.

        A TimeSeriesDataset is a dataset backed by timeseries, but resampled,
        aligned, and (optionally) filtered.

        Parameters
        ----------
        train_start_date: Union[datetime, str]
            Earliest possible point in the dataset (inclusive)
        train_end_date: Union[datetime, str]
            Earliest possible point in the dataset (exclusive)
        tag_list: Sequence[Union[str, Dict, sensor_tag.SensorTag]]
            List of tags to include in the dataset. The elements can be strings,
            dictionaries or SensorTag namedtuples.
        target_tag_list: Sequence[List[Union[str, Dict, sensor_tag.SensorTag]]]
            List of tags to set as the dataset y. These will be treated the same as
            tag_list when fetching and pre-processing (resampling) but will be split
            into the y return from ``.get_data()``
        data_provider: Union[GordoBaseDataProvider, dict]
            A dataprovider which can provide dataframes for tags from train_start_date to train_end_date
            of which can also be a config definition from a data provider's ``.to_dict()`` method.
        resolution: Optional[str]
            The bucket size for grouping all incoming time data (e.g. "10T").
            Available strings come from https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
            **Note**: If this parameter is ``None`` or ``False``, then _no_ aggregation/resampling is applied to the data.
        row_filter: str
            Filter on the rows. Only rows satisfying the filter will be in the dataset.
            See :func:`gordo.machine.dataset.filter_rows.pandas_filter_rows` for
            further documentation of the filter format.
        aggregation_methods
            Aggregation method(s) to use for the resampled buckets. If a single
            resample method is provided then the resulting dataframe will have names
            identical to the names of the series it got in. If several
            aggregation-methods are provided then the resulting dataframe will
            have a multi-level column index, with the series-name as the first level,
            and the aggregation method as the second level.
            See :py:func::`pandas.core.resample.Resampler#aggregate` for more
            information on possible aggregation methods.
        row_filter_buffer_size: int
            Whatever elements are selected for removal based on the ``row_filter``, will also
            have this amount of elements removed fore and aft.
            Default is zero 0
        asset: Optional[str]
            Asset for which the tags are associated with.
        default_asset: Optional[str]
            Asset which will be used if `asset` is not provided and the tag is not
            resolvable to a specific asset.
        n_samples_threshold: int = 0
            The threshold at which the generated DataFrame is considered to have too few rows of data.
        _kwargs
        """
        self.train_start_date = self._validate_dt(train_start_date)
        self.train_end_date = self._validate_dt(train_end_date)

        if self.train_start_date >= self.train_end_date:
            raise ValueError(
                f"train_end_date ({self.train_end_date}) must be after train_start_date ({self.train_start_date})"
            )

        self.tag_list = normalize_sensor_tags(list(tag_list), asset, default_asset)
        self.target_tag_list = (
            normalize_sensor_tags(list(target_tag_list), asset, default_asset)
            if target_tag_list
            else self.tag_list.copy()
        )
        self.resolution = resolution
        self.data_provider = (
            data_provider
            if not isinstance(data_provider, dict)
            else GordoBaseDataProvider.from_dict(data_provider)
        )
        self.row_filter = row_filter
        self.aggregation_methods = aggregation_methods
        self.row_filter_buffer_size = row_filter_buffer_size
        self.asset = asset
        self.n_samples_threshold = n_samples_threshold

        if not self.train_start_date.tzinfo or not self.train_end_date.tzinfo:
            raise ValueError(
                f"Timestamps ({self.train_start_date}, {self.train_end_date}) need to include timezone "
                f"information"
            )

    def to_dict(self):
        params = super().to_dict()
        to_str = lambda dt: str(dt) if not hasattr(dt, "isoformat") else dt.isoformat()
        params["train_start_date"] = to_str(params["train_start_date"])
        params["train_end_date"] = to_str(params["train_end_date"])
        return params

    @staticmethod
    def _validate_dt(dt: Union[str, datetime]) -> datetime:
        dt = dt if isinstance(dt, datetime) else isoparse(dt)
        if dt.tzinfo is None:
            raise ValueError(
                "Must provide an ISO formatted datetime string with timezone information"
            )
        return dt

    def get_data(self) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:

        series_iter: Iterable[pd.Series] = self.data_provider.load_series(
            train_start_date=self.train_start_date,
            train_end_date=self.train_end_date,
            tag_list=list(set(self.tag_list + self.target_tag_list)),
        )

        # Resample if we have a resolution set, otherwise simply join the series.
        if self.resolution:
            data = self.join_timeseries(
                series_iter,
                self.train_start_date,
                self.train_end_date,
                self.resolution,
                aggregation_methods=self.aggregation_methods,
            )
        else:
            data = pd.concat(series_iter, axis=1, join="inner")

        if len(data) <= self.n_samples_threshold:
            raise InsufficientDataError(
                f"The length of the generated DataFrame ({len(data)}) does not exceed the "
                f"specified required threshold for number of rows ({self.n_samples_threshold})."
            )

        if self.row_filter:
            data = pandas_filter_rows(
                data, self.row_filter, buffer_size=self.row_filter_buffer_size
            )

            if len(data) <= self.n_samples_threshold:
                raise InsufficientDataAfterRowFilteringError(
                    f"The length of the genrated DataFrame ({len(data)}) does not exceed the "
                    f"specified required threshold for the number of rows ({self.n_samples_threshold}), "
                    f" after applying the specified row-filter."
                )

        x_tag_names = [tag.name for tag in self.tag_list]
        y_tag_names = [tag.name for tag in self.target_tag_list]

        X = data[x_tag_names]
        y = data[y_tag_names] if self.target_tag_list else None

        if X.first_valid_index():
            self._metadata["train_start_date_actual"] = X.index[0]
            self._metadata["train_end_date_actual"] = X.index[-1]

        self._metadata["summary_statistics"] = X.describe().to_dict()
        hists = dict()
        for tag in X.columns:
            step = round((X[tag].max() - X[tag].min()) / 100, 6)
            if step < 9e-07:
                hists[str(tag)] = "{}"
                continue
            outs = pd.cut(
                X[tag],
                bins=np.arange(
                    round(X[tag].min() - step, 6), round(X[tag].max() + step, 6), step,
                ),
                retbins=False,
            )
            hists[str(tag)] = outs.value_counts().sort_index().to_json(orient="index")
        self._metadata["x_hist"] = hists

        return X, y

    def get_metadata(self):
        return self._metadata.copy()


class RandomDataset(TimeSeriesDataset):
    """
    Get a TimeSeriesDataset backed by
    gordo.machine.dataset.data_provider.providers.RandomDataProvider
    """

    @compat
    @capture_args
    def __init__(
        self,
        train_start_date: Union[datetime, str],
        train_end_date: Union[datetime, str],
        tag_list: list,
        **kwargs,
    ):
        kwargs.pop("data_provider", None)  # Dont care what you ask for, you get random!
        super().__init__(
            data_provider=RandomDataProvider(),
            train_start_date=train_start_date,
            train_end_date=train_end_date,
            tag_list=tag_list,
            **kwargs,
        )



import pytest
import numpy as np
import pandas as pd
from pandas.util.testing import assert_frame_equal
from pandas.core.computation.ops import UndefinedVariableError
from gordo.machine.dataset.filter_rows import pandas_filter_rows, apply_buffer


def test_filter_rows_basic():
    df = pd.DataFrame(list(np.ndindex((10, 2))), columns=["Tag  1", "Tag 2"])
    assert len(pandas_filter_rows(df, "`Tag  1` <= `Tag 2`")) == 3
    assert len(pandas_filter_rows(df, "`Tag  1` == `Tag 2`")) == 2
    assert len(pandas_filter_rows(df, "(`Tag  1` <= `Tag 2`) | `Tag 2` < 2")) == 20
    assert len(pandas_filter_rows(df, "(`Tag  1` <= `Tag 2`) | `Tag 2` < 0.9")) == 12

    assert_frame_equal(
        pandas_filter_rows(df, "(`Tag  1` <= `Tag 2`)"),
        pandas_filter_rows(df, "~(`Tag  1` > `Tag 2`)"),
    )


def test_filter_rows_catches_illegal():
    df = pd.DataFrame(list(np.ndindex((10, 2))), columns=["Tag  1", "Tag 2"])
    with pytest.raises(UndefinedVariableError):
        pandas_filter_rows(df, "sys.exit(0)")
    with pytest.raises(NotImplementedError):
        pandas_filter_rows(df, "lambda x:x")
    with pytest.raises(ValueError):
        pandas_filter_rows(df, "__import__('os').system('clear')"), ValueError


@pytest.mark.parametrize(
    "buffer_size,series,expected",
    [
        (
            1,
            pd.Series([1, 1, 1, 1, 0, 1, 1, 1, 1]),
            pd.Series([1, 1, 1, 0, 0, 0, 1, 1, 1]),
        ),
        (
            2,
            pd.Series([1, 1, 1, 1, 0, 1, 1, 1, 1]),
            pd.Series([1, 1, 0, 0, 0, 0, 0, 1, 1]),
        ),
        (
            1,
            pd.Series([0, 1, 1, 1, 1, 1, 1, 1, 0]),
            pd.Series([0, 0, 1, 1, 1, 1, 1, 0, 0]),
        ),
        (
            2,
            pd.Series([0, 0, 1, 0, 1, 1, 1, 1, 1]),
            pd.Series([0, 0, 0, 0, 0, 0, 1, 1, 1]),
        ),
    ],
)
def test_filter_rows_buffer(buffer_size, series, expected):
    series = series.astype(bool)
    expected = expected.astype(bool)

    apply_buffer(series, buffer_size=buffer_size)
    assert np.alltrue(np.equal(series, expected))



# -*- coding: utf-8 -*-

import logging
import os
import re

import docker
import pytest
import yaml
from unittest.mock import patch

from click.testing import CliRunner

from gordo.workflow.workflow_generator import workflow_generator as wg
from gordo import cli
from gordo.workflow.config_elements.normalized_config import NormalizedConfig
from gordo.machine.dataset import sensor_tag


logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def path_to_config_files():
    """
    Return the data path containing workflow generator test configuration files
    """
    return os.path.join(os.path.dirname(__file__), "data")


def _generate_test_workflow_yaml(
    path_to_config_files, config_filename, project_name="test-proj-name"
):
    """
    Reads a test-config file with workflow_generator, and returns the parsed
    yaml of the generated workflow
    """
    getvalue = _generate_test_workflow_str(
        path_to_config_files, config_filename, project_name=project_name
    )
    expanded_template = yaml.load(getvalue, Loader=yaml.FullLoader)

    return expanded_template


def _generate_test_workflow_str(
    path_to_config_files, config_filename, project_name="test-proj-name"
):
    """
    Reads a test-config file with workflow_generator, and returns the string
    content of the generated workflow
    """
    config_file = os.path.join(path_to_config_files, config_filename)
    args = [
        "workflow",
        "generate",
        "--machine-config",
        config_file,
        "--project-name",
        project_name,
    ]
    runner = CliRunner()

    with patch.object(sensor_tag, "_asset_from_tag_name", return_value="default"):
        result = runner.invoke(cli.gordo, args)

    if result.exception is not None:
        raise result.exception
    return result.output


def _get_env_for_machine_build_serve_task(machine, expanded_template):
    templates = expanded_template["spec"]["templates"]
    do_all = [task for task in templates if task["name"] == "do-all"][0]
    model_builder_machine = [
        task
        for task in do_all["dag"]["tasks"]
        if task["name"] == f"model-builder-{machine}"
    ][0]
    model_builder_machine_env = {
        e["name"]: e["value"] for e in model_builder_machine["arguments"]["parameters"]
    }
    return model_builder_machine_env


@pytest.mark.dockertest
def test_argo_lint(repo_dir, tmpdir, argo_version):
    """
    Test the example config files, assumed to be valid, produces a valid workflow via `argo lint`
    """
    docker_client = docker.from_env()

    # Verify doing a workflow generation on the config will generate valid yaml
    # if it's not, running argo lint will generate wildly unhelpful error msgs.
    config = _generate_test_workflow_yaml(
        path_to_config_files=os.path.join(repo_dir, "examples"),
        config_filename="config.yaml",
    )
    assert isinstance(config, dict)

    workflow_output_path = os.path.join(tmpdir, "out.yml")
    with open(workflow_output_path, "w") as f:
        yaml.dump(config, f)

    logger.info("Running workflow generator and argo lint on examples/config.yaml...")
    result = docker_client.containers.run(
        f"argoproj/argocli:{argo_version}",
        command="lint /tmp/out.yml",
        auto_remove=True,
        stderr=True,
        stdout=True,
        detach=False,
        volumes={str(tmpdir): {"bind": "/tmp", "mode": "ro"}},
    )
    assert result.decode().strip().split("\n")[-1] == "Workflow manifests validated"


def test_basic_generation(path_to_config_files):
    """
    Model must be included in the config file

    start/end dates ...always included? or default to specific dates if not included?
    """

    project_name = "some-fancy-project-name"
    model_config = '{"sklearn.pipeline.Pipeline": {"steps": ["sklearn.preprocessing.data.MinMaxScaler", {"gordo.machine.model.models.KerasAutoEncoder": {"kind": "feedforward_hourglass"}}]}}'

    config_filename = "config-test-with-models.yml"
    expanded_template = _generate_test_workflow_str(
        path_to_config_files, config_filename, project_name=project_name
    )

    assert (
        project_name in expanded_template
    ), f"Expected to find project name: {project_name} in output: {expanded_template}"

    assert (
        model_config in expanded_template
    ), f"Expected to find model config: {model_config} in output: {expanded_template}"

    yaml_content = wg.get_dict_from_yaml(
        os.path.join(path_to_config_files, config_filename)
    )

    with patch.object(sensor_tag, "_asset_from_tag_name", return_value="default"):
        machines = NormalizedConfig(yaml_content, project_name=project_name).machines

    assert len(machines) == 2


def test_generation_to_file(tmpdir, path_to_config_files):
    """
    Test that the workflow generator can output to a file, and it matches
    what would have been output to stdout.
    """
    project_name = "my-sweet-project"
    config_filename = "config-test-with-models.yml"
    expanded_template = _generate_test_workflow_str(
        path_to_config_files, config_filename, project_name=project_name
    )

    # Execute CLI by passing a file to write to.
    config_file = os.path.join(path_to_config_files, config_filename)
    outfile = os.path.join(tmpdir, "out.yml")
    args = [
        "workflow",
        "generate",
        "--machine-config",
        config_file,
        "--project-name",
        project_name,
        "--output-file",
        outfile,
    ]
    runner = CliRunner()
    with patch.object(sensor_tag, "_asset_from_tag_name", return_value="default"):
        result = runner.invoke(cli.gordo, args)
    assert result.exit_code == 0

    # Open the file and ensure they are the same
    with open(outfile, "r") as f:
        outfile_contents = f.read()
    assert outfile_contents.rstrip() == expanded_template.rstrip()


def test_quotes_work(path_to_config_files):
    """Tests that quotes various places result in valid yaml"""
    expanded_template = _generate_test_workflow_yaml(
        path_to_config_files, "config-test-quotes.yml"
    )
    model_builder_machine_1_env = _get_env_for_machine_build_serve_task(
        "machine-1", expanded_template
    )

    machine_1_metadata = yaml.safe_load(model_builder_machine_1_env["machine"])
    assert machine_1_metadata["metadata"]["user_defined"]["machine-metadata"] == {
        "withSingle": "a string with ' in it",
        "withDouble": 'a string with " in it',
        "single'in'key": "why not",
    }

    machine_1_dataset = yaml.safe_load(model_builder_machine_1_env["machine"])
    assert machine_1_dataset["dataset"]["tag_list"] == ["CT/1", 'CT"2', "CT'3"]


def test_overrides_builder_datasource(path_to_config_files):
    expanded_template = _generate_test_workflow_yaml(
        path_to_config_files, "config-test-datasource.yml"
    )

    model_builder_machine_1_env = _get_env_for_machine_build_serve_task(
        "machine-1", expanded_template
    )
    model_builder_machine_2_env = _get_env_for_machine_build_serve_task(
        "machine-2", expanded_template
    )
    model_builder_machine_3_env = _get_env_for_machine_build_serve_task(
        "machine-3", expanded_template
    )

    # ct_23_0002 uses the global overriden requests, but default limits
    assert {"type": "DataLakeProvider", "threads": 20} == yaml.safe_load(
        model_builder_machine_1_env["machine"]
    )["dataset"]["data_provider"]

    # This value must be changed if we change the default values
    assert {"type": "RandomDataProvider", "threads": 15} == yaml.safe_load(
        model_builder_machine_2_env["machine"]
    )["dataset"]["data_provider"]

    # ct_23_0003 uses locally overriden request memory
    assert {"type": "DataLakeProvider", "threads": 10} == yaml.safe_load(
        model_builder_machine_3_env["machine"]
    )["dataset"]["data_provider"]


def test_runtime_overrides_builder(path_to_config_files):
    expanded_template = _generate_test_workflow_yaml(
        path_to_config_files, "config-test-runtime-resource.yaml"
    )
    templates = expanded_template["spec"]["templates"]
    model_builder_task = [
        task for task in templates if task["name"] == "model-builder"
    ][0]
    model_builder_resource = model_builder_task["container"]["resources"]

    # We use yaml overriden memory (both request and limits).
    assert model_builder_resource["requests"]["memory"] == "121M"

    # This was specified to 120 in the config file, but is bumped to match the
    # request
    assert model_builder_resource["limits"]["memory"] == "121M"
    # requests.cpu is all default
    assert model_builder_resource["requests"]["cpu"] == "1001m"


def test_runtime_overrides_client_para(path_to_config_files):
    """
    It is possible to override the parallelization of the client
    through the globals-section of the config file
    """
    expanded_template = _generate_test_workflow_yaml(
        path_to_config_files, "config-test-runtime-resource.yaml"
    )
    templates = expanded_template["spec"]["templates"]
    client_task = [task for task in templates if task["name"] == "gordo-client-waiter"][
        0
    ]

    client_env = {e["name"]: e["value"] for e in client_task["script"]["env"]}

    assert client_env["GORDO_MAX_CLIENTS"] == "10"


def test_runtime_overrides_client(path_to_config_files):
    expanded_template = _generate_test_workflow_yaml(
        path_to_config_files, "config-test-runtime-resource.yaml"
    )
    templates = expanded_template["spec"]["templates"]
    model_client_task = [task for task in templates if task["name"] == "gordo-client"][
        0
    ]
    model_client_resource = model_client_task["script"]["resources"]

    # We use yaml overriden memory (both request and limits).
    assert model_client_resource["requests"]["memory"] == "221M"

    # This was specified to 120 in the config file, but is bumped to match the
    # request
    assert model_client_resource["limits"]["memory"] == "221M"
    # requests.cpu is all default
    assert model_client_resource["requests"]["cpu"] == "100m"


def test_runtime_overrides_influx(path_to_config_files):
    expanded_template = _generate_test_workflow_yaml(
        path_to_config_files, "config-test-runtime-resource.yaml"
    )
    templates = expanded_template["spec"]["templates"]
    influx_task = [
        task for task in templates if task["name"] == "gordo-influx-statefulset"
    ][0]
    influx_statefulset_definition = yaml.load(
        influx_task["resource"]["manifest"], Loader=yaml.FullLoader
    )
    influx_resource = influx_statefulset_definition["spec"]["template"]["spec"][
        "containers"
    ][0]["resources"]
    # We use yaml overriden memory (both request and limits).
    assert influx_resource["requests"]["memory"] == "321M"

    # This was specified to 120 in the config file, but is bumped to match the
    # request
    assert influx_resource["limits"]["memory"] == "321M"
    # requests.cpu is default
    assert influx_resource["requests"]["cpu"] == "520m"
    assert influx_resource["limits"]["cpu"] == "10040m"


def test_disable_influx(path_to_config_files):
    """
    It works to disable influx globally
    """
    expanded_template = _generate_test_workflow_yaml(
        path_to_config_files, "config-test-disable-influx.yml"
    )
    templates = expanded_template["spec"]["templates"]
    do_all = [task for task in templates if task["name"] == "do-all"][0]
    influx_tasks = [
        task["name"] for task in do_all["dag"]["tasks"] if "influx" in task["name"]
    ]
    client_tasks = [
        task["name"] for task in do_all["dag"]["tasks"] if "client" in task["name"]
    ]

    # The cleanup should be the only influx-related task
    assert influx_tasks == ["influx-cleanup"]
    assert client_tasks == []


def test_selective_influx(path_to_config_files):
    """
    It works to enable/disable influx per machine
    """
    expanded_template = _generate_test_workflow_yaml(
        path_to_config_files, "config-test-selective-influx.yml"
    )
    templates = expanded_template["spec"]["templates"]
    do_all = [task for task in templates if task["name"] == "do-all"][0]
    influx_tasks = [
        task["name"] for task in do_all["dag"]["tasks"] if "influx" in task["name"]
    ]
    client_tasks = [
        task["name"] for task in do_all["dag"]["tasks"] if "client" in task["name"]
    ]

    # Now we should have both influx and influx-cleanup
    assert influx_tasks == ["influx-cleanup", "gordo-influx"]

    # And we have a single client task for the one client we want running
    assert client_tasks == ["gordo-client-ct-23-0002"]


@pytest.mark.parametrize("output_to_file", (True, False))
def test_main_tag_list(output_to_file, path_to_config_files, tmpdir):
    config_file = os.path.join(path_to_config_files, "config-test-tag-list.yml")
    args = ["workflow", "unique-tags", "--machine-config", config_file]

    out_file = os.path.join(tmpdir, "out.txt")

    if output_to_file:
        args.extend(["--output-file-tag-list", out_file])

    runner = CliRunner()
    with patch.object(sensor_tag, "_asset_from_tag_name", return_value="default"):
        result = runner.invoke(cli.gordo, args)

    assert result.exit_code == 0

    if output_to_file:
        assert os.path.isfile(out_file)
    else:
        output_tags = set(result.output.split(sep="\n")[:-1])
        expected_output_tags = {"Tag 1", "Tag 2", "Tag 3", "Tag 4", "Tag 5"}

        assert (
            output_tags == expected_output_tags
        ), f"Expected to find: {expected_output_tags}, outputted {output_tags}"


def test_valid_dateformats(path_to_config_files):
    output_workflow = _generate_test_workflow_str(
        path_to_config_files, "config-test-allowed-timestamps.yml"
    )
    # Three from the dataset, three from the start for tag fetching, and three in
    # each machine's model-crd specification
    assert output_workflow.count("2016-11-07") == 9
    assert output_workflow.count("2017-11-07") == 6


def test_model_names_embedded(path_to_config_files):
    """
    Tests that the generated workflow contains the names of the machines
    it builds a workflow for in the metadata/annotation as a yaml-parsable structure
    """
    output_workflow = _generate_test_workflow_yaml(
        path_to_config_files, "config-test-allowed-timestamps.yml"
    )
    parsed_machines = yaml.load(
        output_workflow["metadata"]["annotations"]["gordo-models"],
        Loader=yaml.FullLoader,
    )
    assert parsed_machines == ["machine-1", "machine-2", "machine-3"]


def test_missing_timezone(path_to_config_files):
    with pytest.raises(ValueError):
        _generate_test_workflow_yaml(
            path_to_config_files, "config-test-missing-timezone.yml"
        )

    with pytest.raises(ValueError):
        _generate_test_workflow_yaml(
            path_to_config_files, "config-test-missing-timezone-quoted.yml"
        )


def test_validates_resource_format(path_to_config_files):
    """
    We validate that resources are integers
    """
    with pytest.raises(ValueError):
        _generate_test_workflow_str(
            path_to_config_files, "config-test-failing-resource-format.yml"
        )


@pytest.mark.parametrize(
    "owner_ref_str,valid",
    (
        ("[]", False),
        ("- key: value", False),
        (
            """
            - uid: 1
              name: name
              kind: kind
              apiVersion: v1
            """,
            True,
        ),
    ),
)
def test_valid_owner_ref(owner_ref_str: str, valid: bool):
    if valid:
        wg._valid_owner_ref(owner_ref_str)
    else:
        with pytest.raises(TypeError):
            wg._valid_owner_ref(owner_ref_str)


@pytest.mark.parametrize(
    "test_file, log_level",
    (
        ("config-test-with-log-key.yml", "DEBUG"),
        ("config-test-with-models.yml", "INFO"),
    ),
)
def test_log_level_key(test_file: str, log_level: str, path_to_config_files: str):
    """
    Test that GORDO_LOG_LEVEL is set to the correct value if specified in the config file, or default to INFO if not
    specified.
    """
    workflow_str = _generate_test_workflow_str(path_to_config_files, test_file)

    # Find the value on the next line after the key GORDO_LOG_LEVEL
    gordo_log_levels = re.findall(
        r"(?<=GORDO_LOG_LEVEL\r|GORDO_LOG_LEVEL\n)[^\r\n]+", workflow_str
    )

    # Assert all the values to the GORDO_LOG_LEVEL key contains the correct log-level
    assert all([log_level in value for value in gordo_log_levels])


def test_expected_models_in_workflow(repo_dir):
    """
    Server deployment depends on EXPECTED_MODELS env var being set,
    which is a list of strings, indicating the expected model names to be served.
    """
    workflow_str = _generate_test_workflow_str(
        path_to_config_files=os.path.join(repo_dir, "examples"),
        config_filename="config.yaml",
    )
    assert "name: EXPECTED_MODELS" in workflow_str

    # Not the prettiest, but a whole lot prettier than digging down into the workflow yaml
    # basically want to get to 'gordo-server-deployment' and ensure the EXPECTED_MODELS env var
    # is set with a list (in string form) of model names which can be loaded.
    expected_models_str = (
        workflow_str.split("EXPECTED_MODELS")[1]
        .split("value:")[1]
        .split("|")[1]
        .split()[0]
    )
    assert isinstance(yaml.safe_load(expected_models_str), list)



# -*- coding: utf-8 -*-

from datetime import timedelta

import pytest
import numpy as np
import pandas as pd
import yaml

from sklearn.decomposition import PCA
from sklearn.preprocessing import MinMaxScaler, RobustScaler
from sklearn.multioutput import MultiOutputRegressor
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import cross_validate
from sklearn.model_selection import TimeSeriesSplit

from gordo import serializer
from gordo.machine.model import utils as model_utils
from gordo.machine.model.anomaly import DiffBasedAnomalyDetector
from gordo.machine.model.anomaly.base import AnomalyDetectorBase


@pytest.mark.parametrize("scaler", (MinMaxScaler(), RobustScaler()))
@pytest.mark.parametrize(
    "index", (range(10), pd.date_range("2019-01-01", "2019-01-30", periods=10))
)
@pytest.mark.parametrize("lookback", (0, 5))
@pytest.mark.parametrize("with_thresholds", (True, False))
def test_diff_detector(scaler, index, lookback, with_thresholds: bool):
    """
    Test the functionality of the DiffBasedAnomalyDetector
    """

    # Some dataset.
    X, y = (
        pd.DataFrame(np.random.random((10, 3))),
        pd.DataFrame(np.random.random((10, 3))),
    )

    base_estimator = MultiOutputRegressor(estimator=LinearRegression())
    model = DiffBasedAnomalyDetector(
        base_estimator=base_estimator, scaler=scaler, require_thresholds=False
    )

    assert isinstance(model, AnomalyDetectorBase)

    assert model.get_params() == dict(base_estimator=base_estimator, scaler=scaler)

    if with_thresholds:
        model.cross_validate(X=X, y=y)

    model.fit(X, y)

    output: np.ndarray = model.predict(X)
    base_df = model_utils.make_base_dataframe(
        tags=["A", "B", "C"], model_input=X, model_output=output, index=index
    )

    # Base prediction dataframe has none of these columns
    assert not any(
        col in base_df.columns
        for col in (
            "total-anomaly-scaled",
            "total-anomaly-unscaled",
            "tag-anomaly-scaled",
            "tag-anomaly-unscaled",
        )
    )

    # Apply the anomaly detection logic on the base prediction df
    anomaly_df = model.anomaly(X, y, timedelta(days=1))

    # Should have these added error calculated columns now.
    assert all(
        col in anomaly_df.columns
        for col in (
            "total-anomaly-scaled",
            "total-anomaly-unscaled",
            "tag-anomaly-scaled",
            "tag-anomaly-unscaled",
        )
    )

    # Verify calculation for unscaled data
    feature_error_unscaled = np.abs(base_df["model-output"].values - y.values)
    total_anomaly_unscaled = np.square(feature_error_unscaled).mean(axis=1)
    assert np.allclose(
        feature_error_unscaled, anomaly_df["tag-anomaly-unscaled"].values
    )
    assert np.allclose(
        total_anomaly_unscaled, anomaly_df["total-anomaly-unscaled"].values
    )

    # Verify calculations for scaled data
    feature_error_scaled = np.abs(
        scaler.transform(base_df["model-output"].values) - scaler.transform(y)
    )
    total_anomaly_scaled = np.square(feature_error_scaled).mean(axis=1)
    assert np.allclose(feature_error_scaled, anomaly_df["tag-anomaly-scaled"].values)
    assert np.allclose(total_anomaly_scaled, anomaly_df["total-anomaly-scaled"].values)

    if with_thresholds:
        assert "anomaly-confidence" in anomaly_df.columns
        assert "total-anomaly-confidence" in anomaly_df.columns
    else:
        assert "anomaly-confidence" not in anomaly_df.columns
        assert "total-anomaly-confidence" not in anomaly_df.columns


@pytest.mark.parametrize(
    "config",
    (
        """
        gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
            base_estimator:
                sklearn.pipeline.Pipeline:
                    steps:
                        - sklearn.preprocessing.data.MinMaxScaler
                        - gordo.machine.model.models.KerasAutoEncoder:
                            kind: feedforward_hourglass
                    memory:
    """,
        """
        gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
            base_estimator:
                gordo.machine.model.models.KerasAutoEncoder:
                    kind: feedforward_hourglass
    """,
    ),
)
def test_diff_detector_serializability(config):
    """
    Should play well with the gordo serializer
    """
    config = yaml.load(config)

    model = serializer.from_definition(config)
    serializer.into_definition(model)
    serialized_bytes = serializer.dumps(model)
    serializer.loads(serialized_bytes)


@pytest.mark.parametrize("n_features_y", range(1, 3))
@pytest.mark.parametrize("n_features_x", range(1, 3))
def test_diff_detector_threshold(n_features_y: int, n_features_x: int):
    """
    Basic construction logic of thresholds_ attribute in the
    DiffBasedAnomalyDetector
    """
    X = np.random.random((100, n_features_x))
    y = np.random.random((100, n_features_y))

    model = DiffBasedAnomalyDetector(
        base_estimator=MultiOutputRegressor(estimator=LinearRegression())
    )

    # Model has own implementation of cross_validate
    assert hasattr(model, "cross_validate")

    # When initialized it should not have a threshold calculated.
    assert not hasattr(model, "feature_thresholds_")
    assert not hasattr(model, "aggregate_threshold_")
    assert not hasattr(model, "feature_thresholds_per_fold_")
    assert not hasattr(model, "aggregate_thresholds_per_fold_")

    model.fit(X, y)

    # Until it has done cross validation, it has no threshold.
    assert not hasattr(model, "feature_thresholds_")
    assert not hasattr(model, "aggregate_threshold_")
    assert not hasattr(model, "feature_thresholds_per_fold_")
    assert not hasattr(model, "aggregate_thresholds_per_fold_")

    # Calling cross validate should set the threshold for it.
    model.cross_validate(X=X, y=y)

    # Now we have calculated thresholds based on cross validation folds
    assert hasattr(model, "feature_thresholds_")
    assert hasattr(model, "aggregate_threshold_")
    assert hasattr(model, "feature_thresholds_per_fold_")
    assert hasattr(model, "aggregate_thresholds_per_fold_")
    assert isinstance(model.feature_thresholds_, pd.Series)
    assert len(model.feature_thresholds_) == y.shape[1]
    assert all(model.feature_thresholds_.notna())
    assert isinstance(model.feature_thresholds_per_fold_, pd.DataFrame)
    assert isinstance(model.aggregate_thresholds_per_fold_, dict)


@pytest.mark.parametrize("return_estimator", (True, False))
def test_diff_detector_cross_validate(return_estimator: bool):
    """
    DiffBasedAnomalyDetector.cross_validate implementation should be the
    same as sklearn.model_selection.cross_validate if called the same.

    And it always will update `return_estimator` to True, as it requires
    the intermediate models to calculate the thresholds
    """
    X = np.random.random((100, 10))
    y = np.random.random((100, 1))

    model = DiffBasedAnomalyDetector(base_estimator=LinearRegression())

    cv = TimeSeriesSplit(n_splits=3)
    cv_results_da = model.cross_validate(
        X=X, y=y, cv=cv, return_estimator=return_estimator
    )
    cv_results_sk = cross_validate(model, X=X, y=y, cv=cv, return_estimator=True)

    assert cv_results_da.keys() == cv_results_sk.keys()


@pytest.mark.parametrize("require_threshold", (True, False))
def test_diff_detector_require_thresholds(require_threshold: bool):
    """
    Should fail if requiring thresholds, but not calling cross_validate
    """
    X = pd.DataFrame(np.random.random((100, 5)))
    y = pd.DataFrame(np.random.random((100, 2)))

    model = DiffBasedAnomalyDetector(
        base_estimator=MultiOutputRegressor(LinearRegression()),
        require_thresholds=require_threshold,
    )

    model.fit(X, y)

    if require_threshold:
        # FAIL: Forgot to call .cross_validate to calculate thresholds.
        with pytest.raises(AttributeError):
            model.anomaly(X, y)

        model.cross_validate(X=X, y=y)
        model.anomaly(X, y)
    else:
        # thresholds not required
        model.anomaly(X, y)



import pytest
import yaml
import tensorflow as tf
import numpy as np
from sklearn.pipeline import Pipeline

from gordo import serializer
from gordo.machine.model.models import KerasRawModelRegressor


@pytest.mark.parametrize(
    "spec_str",
    (
        """
        compile:
            loss: mse
            optimizer: 
              tensorflow.keras.optimizers.SGD:
                learning_rate: 0.001
        spec:
            tensorflow.keras.models.Sequential:
                layers:
                    - tensorflow.keras.layers.Dense: 
                        units: 10
                    - tensorflow.keras.layers.Dense:
                        units: 32
                        kernel_regularizer:
                            tensorflow.keras.regularizers.L1L2:
                                l1: 0.2
                    - tensorflow.keras.layers.Dense:
                        units: 1
    """,
        """
        compile:
            loss: mse
            optimizer: adam
        spec:
            tensorflow.keras.models.Sequential:
                layers:
                    - tensorflow.keras.layers.Input:
                        shape: [9]
                    - tensorflow.keras.layers.Reshape:
                        target_shape: [3, 3]
                    - tensorflow.keras.layers.LSTM:
                        units: 12
                    - tensorflow.keras.layers.Flatten
                    - tensorflow.keras.layers.Dense:
                        units: 1
    """,
    ),
)
def test_raw_keras_basic(spec_str: str):
    """
    Can load a keras.Sequential model from a config/yaml definition(s)
    """
    spec = yaml.safe_load(spec_str)
    pipe = KerasRawModelRegressor(spec)
    model = pipe.build_fn()
    assert isinstance(model, tf.keras.models.Sequential)


def test_raw_keras_part_of_pipeline():
    """
    It should play well, when tucked into a sklearn.pipeline.Pipeline
    """
    X, y = np.random.random((100, 4)), np.random.random((100, 1))

    config_str = """
    sklearn.pipeline.Pipeline:
        steps:
            - sklearn.decomposition.pca.PCA:
                n_components: 4
            - gordo.machine.model.models.KerasRawModelRegressor:
                kind:
                    compile:
                        loss: mse
                        optimizer: adam
                    spec:
                        tensorflow.keras.models.Sequential:
                            layers:
                                - tensorflow.keras.layers.Dense:
                                    units: 4
                                - tensorflow.keras.layers.Dense:
                                    units: 1
    """
    config = yaml.safe_load(config_str)
    pipe = serializer.from_definition(config)
    assert isinstance(pipe, Pipeline)

    pipe.fit(X, y)
    out = pipe.predict(X)
    assert len(out) == len(y)



# -*- coding: utf-8 -*-

import abc
import itertools
import logging
import time
from typing import Any, Dict, Optional

import pandas as pd
import numpy as np

from gordo.client.utils import influx_client_from_uri
from gordo.machine import Machine


logger = logging.getLogger(__name__)


class PredictionForwarder(metaclass=abc.ABCMeta):

    """
    Definition of a callable which the :class:`gordo.client.Client`
    will call after each successful prediction response::

        def my_forwarder(
            predictions: pd.DataFrame = None,
            machine: Machine = None,
            metadata: dict = dict(),
            resampled_sensor_data: pd.DataFrame = None
        ):
            ...
    """

    @abc.abstractmethod
    def __call__(
        self,
        *,
        predictions: pd.DataFrame = None,
        machine: Machine = None,
        metadata: dict = dict(),
        resampled_sensor_data: pd.DataFrame = None,
    ):
        ...


class ForwardPredictionsIntoInflux(PredictionForwarder):
    """
    To be used as a 'forwarder' for the prediction client

    After instantiation, it is a coroutine which accepts prediction dataframes
    which it will pass onto influx
    """

    def __init__(
        self,
        destination_influx_uri: Optional[str] = None,
        destination_influx_api_key: Optional[str] = None,
        destination_influx_recreate: bool = False,
        n_retries=5,
    ):
        """
        Create an instance which, when called, is a coroutine capable of
        being sent dataframes generated from the '/anomaly/prediction' machine

        Parameters
        ----------
        destination_influx_uri: str
            Connection string for destination influx -
            format: <username>:<password>@<host>:<port>/<optional-path>/<db_name>
        destination_influx_api_key: str
            API key if needed for destination db
        destination_influx_recreate: bool
            Drop the database before filling it with data?
        """
        # Create df client if provided
        self.n_retries = n_retries
        self.dataframe_client = (
            influx_client_from_uri(
                destination_influx_uri,
                api_key=destination_influx_api_key,
                recreate=destination_influx_recreate,
                dataframe_client=True,
            )
            if destination_influx_uri
            else None
        )

    def __call__(
        self,
        *,
        predictions: pd.DataFrame = None,
        machine: Machine = None,
        metadata: dict = dict(),
        resampled_sensor_data: pd.DataFrame = None,
    ):
        # clean predictions for possible inf, nan, which influx can't handle
        if predictions is not None:
            predictions = self._clean_df(predictions)
        if resampled_sensor_data is not None:
            resampled_sensor_data = self._clean_df(resampled_sensor_data)

        if resampled_sensor_data is None and predictions is None:
            raise ValueError(
                "Argument `resampled_sensor_data` or `predictions` must be passed"
            )
        if predictions is not None:
            if machine is None:
                raise ValueError(
                    "Argument `machine`must be provided if `predictions` is provided"
                )
            self.forward_predictions(predictions, machine=machine, metadata=metadata)
        if resampled_sensor_data is not None:
            self.send_sensor_data(resampled_sensor_data)

    @staticmethod
    def _clean_df(df: pd.DataFrame) -> pd.DataFrame:
        """
        Ensure dataframe doesn't have inf / nan values which influx can't handle

        Parameters
        ----------
        df: pandas.DataFrame

        Returns
        -------
        pandas.DataFrame
        """
        return df.replace([np.inf, -np.inf], np.nan).dropna()

    def forward_predictions(
        self, predictions: pd.DataFrame, machine: Machine, metadata: dict = dict()
    ):
        """
        Takes a multi-layed column dataframe and write points to influx where
        each top level name is treated as the measurement name.

        How the data is written via InfluxDB DataFrameClient in this method
        determines the schema of the database.

        Parameters
        ----------
        predictions: pd.DataFrame
            Multi layed column dataframe, where top level names will be treated
            as the 'measurement' name in influx and 2nd level will be the fields
            under those measurements.

        Returns
        -------
        None
        """
        # Setup tags; metadata (if any) and other key value pairs.
        tags = {"machine": f"{machine.name}"}
        tags.update(metadata)

        # The measurements to be posted to Influx
        top_lvl_names = predictions.columns.get_level_values(0).unique()

        for top_lvl_name in top_lvl_names:

            # Makes no sense to create a influx point where the field would be a date
            # and doesn't seem to be possible anyway.
            if top_lvl_name in ["end", "start"]:
                continue

            # this is a 'regular' non-stacked column dataframe
            sub_df = predictions[top_lvl_name]

            if isinstance(sub_df, pd.Series):
                sub_df = pd.DataFrame(sub_df)

            # Set the sub df's column names equal to the name of the tags if
            # they match the length of the tag list.
            if len(sub_df.columns) == len(machine.dataset.tag_list):
                sub_df.columns = [tag.name for tag in machine.dataset.tag_list]

            self._write_to_influx_with_retries(sub_df, top_lvl_name, tags)

    def _write_to_influx_with_retries(self, df, measurement, tags: Dict[str, Any] = {}):
        """
        Write data to influx with retries and exponential backof. Will sleep
        exponentially longer between each retry, starting at 8 seconds, capped at 5 min.
        """
        logger.info(
            f"Writing {len(df)} points to Influx for measurement: {measurement}"
        )

        for current_attempt in itertools.count(start=1):
            try:
                df = ForwardPredictionsIntoInflux._stack_to_name_value_columns(df)

                self.dataframe_client.write_points(
                    dataframe=df,
                    measurement=measurement,
                    tags=tags,
                    tag_columns=["sensor_name"],
                    field_columns=["sensor_value"],
                    batch_size=10000,
                )
            except Exception as exc:
                if current_attempt <= self.n_retries:
                    # Sleep at most 5 min
                    time_to_sleep = min(2 ** (current_attempt + 2), 300)
                    logger.warning(
                        f"Failed to forward data to influx on attempt "
                        f"{current_attempt} out of {self.n_retries}.\n"
                        f"Error: {exc}.\n"
                        f"Sleeping {time_to_sleep} seconds and trying again."
                    )
                    time.sleep(time_to_sleep)
                    continue
                else:
                    msg = f"Failed to forward data to influx. Error: {exc}"
                    logger.error(msg)
            else:
                break

    def send_sensor_data(self, sensors: pd.DataFrame):
        """
        Write sensor-data to influx
        """
        # Write the per-sensor points to influx
        logger.info(f"Writing {len(sensors)} sensor points to Influx")
        self._write_to_influx_with_retries(sensors, "resampled")
        logger.debug("Done writing resampled sensor values to influx")

    @staticmethod
    def _stack_to_name_value_columns(df: pd.DataFrame) -> pd.DataFrame:
        """
        Stack a DataFrame with a range of tag columns to one with name and value columns

        Parameters
        ----------
        df: pd.DataFrame
            Source dataframe with individual columns for tags.

        Returns
        -------
        df: pd.DataFrame
            Stacked dataframe with columns `sensor_name` and `sensor_value`.
        """
        # String column names are necessary for stacking
        # (as opposed to integers when df created from np.ndarray)
        df.columns = df.columns.astype(str)

        df = df.stack().to_frame(name="sensor_value")
        df = df.reset_index(level=1).rename(columns={"level_1": "sensor_name"})
        df["sensor_value"].astype(float)
        return df



# -*- coding: utf-8 -*-

from typing import List

from gordo.machine.validators import fix_runtime
from gordo.workflow.workflow_generator.helpers import patch_dict
from gordo.machine import Machine


def _calculate_influx_resources(nr_of_machines):
    return {
        "requests": {
            # The requests must be limited to keep the machine schedulable
            "memory": min(3000 + (220 * nr_of_machines), 28000),  # Between 3G and 28G
            "cpu": min(500 + (10 * nr_of_machines), 4000),  # Between 500m and 4000m
        },
        "limits": {
            "memory": min(3000 + (220 * nr_of_machines), 48000),
            "cpu": 10000 + (20 * nr_of_machines),
        },
    }


class NormalizedConfig:

    """
    Handles the conversion of a single Machine representation in config format
    and updates it with any features which are 'left out' inside of ``globals``
    key or the default config globals held here.
    """

    DEFAULT_CONFIG_GLOBALS = {
        "runtime": {
            "reporters": [],
            "server": {
                "resources": {
                    "requests": {"memory": 3000, "cpu": 1000},
                    "limits": {"memory": 6000, "cpu": 2000},
                }
            },
            "builder": {
                "resources": {
                    "requests": {"memory": 3900, "cpu": 1001},
                    "limits": {"memory": 3900, "cpu": 1001},
                },
                "remote_logging": {"enable": False},
            },
            "client": {
                "resources": {
                    "requests": {"memory": 3500, "cpu": 100},
                    "limits": {"memory": 4000, "cpu": 2000},
                },
                "max_instances": 30,
            },
            "influx": {"enable": True},
        },
        "evaluation": {
            "cv_mode": "full_build",
            "scoring_scaler": "sklearn.preprocessing.RobustScaler",
            "metrics": [
                "explained_variance_score",
                "r2_score",
                "mean_squared_error",
                "mean_absolute_error",
            ],
        },
    }

    """
    Represents a fully loaded config file
    """

    machines: List[Machine]
    globals: dict

    def __init__(self, config: dict, project_name: str):
        default_globals = self.DEFAULT_CONFIG_GLOBALS
        default_globals["runtime"]["influx"][  # type: ignore
            "resources"
        ] = _calculate_influx_resources(  # type: ignore
            len(config["machines"])
        )

        passed_globals = config.get("globals", dict())
        patched_globals = patch_dict(default_globals, passed_globals)
        if patched_globals.get("runtime"):
            patched_globals["runtime"] = fix_runtime(patched_globals.get("runtime"))
        self.project_name = project_name
        self.machines = [
            Machine.from_config(
                conf, project_name=project_name, config_globals=patched_globals
            )
            for conf in config["machines"]
        ]  # type: List[Machine]

        self.globals = patched_globals



# -*- coding: utf-8 -*-

import os
import logging
import pytest
import subprocess
import shutil
import json

from typing import List
from unittest.mock import patch

from gordo.server.server import run_cmd
from gordo import serializer, __version__
from gordo.server import server

import tests.utils as tu


logger = logging.getLogger(__name__)


def test_healthcheck_endpoint(base_route, gordo_ml_server_client):
    """
    Test expected behavior of /<gordo-name>/healthcheck
    """
    # Should also be at the very lowest level as well.
    resp = gordo_ml_server_client.get(f"/healthcheck")
    assert resp.status_code == 200

    resp = gordo_ml_server_client.get(f"{base_route}/healthcheck")
    assert resp.status_code == 200

    data = resp.get_json()
    logger.debug(f"Got resulting JSON response: {data}")
    assert "gordo-server-version" in data


def test_response_header_timing(base_route, gordo_ml_server_client):
    """
    Test that the response contains a `Server-Timing` header
    """
    resp = gordo_ml_server_client.get(f"{base_route}/healthcheck")
    assert resp.status_code == 200
    assert "Server-Timing" in resp.headers
    assert "request_walltime_s" in resp.headers["Server-Timing"]


def test_metadata_endpoint(base_route, gordo_single_target, gordo_ml_server_client):
    """
    Test the expected behavior of /metadata
    """
    resp = gordo_ml_server_client.get(f"{base_route}/metadata")
    assert resp.status_code == 200

    data = resp.get_json()
    assert "metadata" in data
    assert data["metadata"]["name"] == gordo_single_target


def test_download_model(api_version, gordo_project, gordo_name, gordo_ml_server_client):
    """
    Test we can download a model, loadable via serializer.loads()
    """
    resp = gordo_ml_server_client.get(
        f"/gordo/{api_version}/{gordo_project}/{gordo_name}/download-model"
    )

    serialized_model = resp.get_data()
    model = serializer.loads(serialized_model)

    # All models have a fit method
    assert hasattr(model, "fit")

    # Models MUST have either predict or transform
    assert hasattr(model, "predict") or hasattr(model, "transform")

    # Asking for a model that doesn't exist gives 404
    resp = gordo_ml_server_client.get(
        f"/gordo/{api_version}/{gordo_project}/invalid-model-name/download-model"
    )
    assert resp.status_code == 404


def test_run_cmd(monkeypatch):
    """
    Test that execution error catchings work as expected
    """

    # Call command that raises FileNotFoundError, a subclass of OSError
    cmd = ["gumikorn", "gordo.server.server:app"]
    with pytest.raises(OSError):
        run_cmd(cmd)

    # Call command that raises a CalledProcessError
    cmd = ["ping", "--bad-option"]
    with pytest.raises(subprocess.CalledProcessError):
        run_cmd(cmd)


@pytest.mark.parametrize("revisions", [("1234", "2345", "3456"), ("1234",)])
def test_list_revisions(tmpdir, revisions: List[str]):
    """
    Verify the server is capable of returning the project revisions
    it's capable of serving.
    """

    # Server gets the 'latest' directory to serve models from, but knows other
    # revisions should be available a step up from this directory.
    model_dir = os.path.join(tmpdir, revisions[0])

    # Make revision directories under the tmpdir
    [os.mkdir(os.path.join(tmpdir, rev)) for rev in revisions]  # type: ignore

    # Request from the server what revisions it can serve, should match
    with tu.temp_env_vars(MODEL_COLLECTION_DIR=model_dir):
        app = server.build_app()
        app.testing = True
        client = app.test_client()
        resp = client.get("/gordo/v0/test-project/revisions")
        resp_with_revision = client.get(
            f"/gordo/v0/test-project/revisions?revision={revisions[-1]}"
        )

    assert set(resp.json.keys()) == {"latest", "available-revisions", "revision"}
    assert resp.json["latest"] == os.path.basename(model_dir)
    assert resp.json["revision"] == os.path.basename(model_dir)
    assert isinstance(resp.json["available-revisions"], list)
    assert set(resp.json["available-revisions"]) == set(revisions)

    # And the request asking to use a specific revision gives back that revision,
    # but will return the expected latest available
    assert resp_with_revision.json["latest"] == os.path.basename(model_dir)
    assert resp_with_revision.json["revision"] == revisions[-1]


def test_list_revisions_listdir_fail(caplog):
    """
    Verify the server will not fail if listing directories above the current
    model collection directory it has, fails.
    """

    def listdir_fail(*args, **kwargs):
        raise FileNotFoundError()

    expected_revision = "some-project-revision-123"

    with patch.object(os, "listdir", side_effect=listdir_fail) as mocked_listdir:
        with caplog.at_level(logging.CRITICAL):
            with tu.temp_env_vars(MODEL_COLLECTION_DIR=expected_revision):
                app = server.build_app()
                app.testing = True
                client = app.test_client()
                resp = client.get("/gordo/v0/test-project/revisions")

    assert mocked_listdir.called_once()
    assert set(resp.json.keys()) == {"latest", "available-revisions", "revision"}
    assert resp.json["latest"] == expected_revision
    assert isinstance(resp.json["available-revisions"], list)
    assert resp.json["available-revisions"] == [expected_revision]


def test_model_list_view_non_existant_proj():
    with tu.temp_env_vars(MODEL_COLLECTION_DIR=os.path.join("does", "not", "exist")):
        app = server.build_app()
        app.testing = True
        client = app.test_client()
        resp = client.get("/gordo/v0/test-project/models")
        assert resp.status_code == 200
        assert resp.json["models"] == []


@pytest.mark.parametrize(
    "revision_to_models",
    [
        {"123": ("model-1", "model-2"), "456": ("model-3", "model-4")},
        {"123": (), "456": ("model-1",)},
        dict(),
    ],
)
def test_models_by_revision_list_view(caplog, tmpdir, revision_to_models):
    """
    Server returns expected models it can serve under specific revisions.

    revision_to_models: Dict[str, Tuple[str, ...]]
        Map revision codes to models belonging to that revision.
        Simulate serving some revision, but having access to other
        revisions and its models.
    """

    # Current collection dir for the server, order isn't important.
    if revision_to_models:
        collection_dir = os.path.join(tmpdir, list(revision_to_models.keys())[0])
    else:
        # This will cause a failure to look up a certain revision
        collection_dir = str(tmpdir)

    # Make all the revision and model subfolders
    for revision in revision_to_models.keys():
        os.mkdir(os.path.join(tmpdir, revision))
        for model in revision_to_models[revision]:
            os.makedirs(os.path.join(tmpdir, revision, model), exist_ok=True)

    with tu.temp_env_vars(MODEL_COLLECTION_DIR=collection_dir):
        app = server.build_app()
        app.testing = True
        client = app.test_client()
        for revision in revision_to_models:
            resp = client.get(f"/gordo/v0/test-project/models?revision={revision}")
            assert resp.status_code == 200
            assert "models" in resp.json
            assert sorted(resp.json["models"]) == sorted(revision_to_models[revision])
        else:
            # revision_to_models is empty, so there is nothing on the server.
            # Test that asking for some arbitrary revision will give a 404 and error message
            resp = client.get(
                f"/gordo/v0/test-project/models?revision=revision-does-not-exist"
            )
            assert resp.status_code == 410
            assert resp.json == {
                "error": "Revision 'revision-does-not-exist' not found.",
                "revision": "revision-does-not-exist",
            }


@pytest.mark.parametrize("revisions", (("123", "456"), ("123",)))
def test_request_specific_revision(trained_model_directory, tmpdir, revisions):

    model_name = "test-model"
    current_revision = revisions[0]
    collection_dir = os.path.join(tmpdir, current_revision)

    # Copy trained model into revision model folders
    for revision in revisions:
        model_dir = os.path.join(tmpdir, revision, model_name)
        shutil.copytree(trained_model_directory, model_dir)

        # Now overwrite the metadata.json file to ensure the server actually reads
        # the metadata for this specific revision
        metadata_file = os.path.join(model_dir, "metadata.json")
        assert os.path.isfile(metadata_file)
        with open(metadata_file, "w") as fp:
            json.dump({"revision": revision, "model": model_name}, fp)

    with tu.temp_env_vars(MODEL_COLLECTION_DIR=collection_dir):
        app = server.build_app()
        app.testing = True
        client = app.test_client()
        for revision in revisions:
            resp = client.get(
                f"/gordo/v0/test-project/{model_name}/metadata?revision={revision}"
            )
            assert resp.status_code == 200
            assert resp.json["revision"] == revision

            # Verify the server read the metadata.json file we had overwritten
            assert resp.json["metadata"] == {"revision": revision, "model": model_name}

        # Asking for a revision which doesn't exist gives a 410 Gone.
        resp = client.get(
            f"/gordo/v0/test-project/{model_name}/metadata?revision=does-not-exist"
        )
        assert resp.status_code == 410
        assert resp.json == {
            "error": "Revision 'does-not-exist' not found.",
            "revision": "does-not-exist",
        }

        # Again but by setting header, to ensure we also check the header
        resp = client.get(
            f"/gordo/v0/test-project/{model_name}/metadata",
            headers={"revision": "does-not-exist"},
        )
        assert resp.status_code == 410
        assert resp.json == {
            "error": "Revision 'does-not-exist' not found.",
            "revision": "does-not-exist",
        }


def test_server_version_route(model_collection_directory, gordo_revision):
    """
    Simple route which returns the current version
    """
    with tu.temp_env_vars(MODEL_COLLECTION_DIR=model_collection_directory):
        app = server.build_app()
        app.testing = True
        client = app.test_client()

        resp = client.get("/server-version")
        assert resp.status_code == 200
        assert resp.json == {"revision": gordo_revision, "version": __version__}


def test_non_existant_model_metadata(tmpdir, gordo_project, api_version):
    """
    Simple route which returns the current version
    """
    with tu.temp_env_vars(MODEL_COLLECTION_DIR=str(tmpdir)):
        app = server.build_app()
        app.testing = True
        client = app.test_client()

        resp = client.get(
            f"/gordo/{api_version}/{gordo_project}/model-does-not-exist/metadata"
        )
        assert resp.status_code == 404


def test_expected_models_route(tmpdir):
    """
    Route that gives back the expected models names, which are just read from
    the 'EXPECTED_MODELS' env var.
    """
    with tu.temp_env_vars(
        MODEL_COLLECTION_DIR=str(tmpdir),
        EXPECTED_MODELS=json.dumps(["model-a", "model-b"]),
    ):
        app = server.build_app()
        app.testing = True
        client = app.test_client()

        resp = client.get("/gordo/v0/test-project/expected-models")
        assert resp.json["expected-models"] == ["model-a", "model-b"]



from gordo.machine import metadata as m


def test_metadata_dataclass():

    # Should be able to create it without arguments
    m.Metadata.from_dict(m.Metadata())

    metadata = m.Metadata(
        user_defined={"some-key": "some-value"},
        build_metadata=m.BuildMetadata(
            model=m.ModelBuildMetadata(
                model_offset=0,
                model_creation_date="2016-01-01",
                model_builder_version="v1",
                cross_validation=m.CrossValidationMetaData(),
            ),
            dataset=m.DatasetBuildMetadata(
                query_duration_sec=1, dataset_meta=dict(key="value")
            ),
        ),
    )
    m.Metadata.from_dict(metadata.to_dict())



# -*- coding: utf-8 -*-

import pickle
import logging
import pydoc

import pytest
import numpy as np

from sklearn.exceptions import NotFittedError
from sklearn.pipeline import Pipeline
from sklearn.model_selection import cross_val_score, TimeSeriesSplit

from tensorflow.keras.wrappers.scikit_learn import KerasRegressor as BaseWrapper

from tests.utils import get_model
from gordo.machine.model.models import (
    KerasLSTMAutoEncoder,
    KerasLSTMForecast,
    KerasLSTMBaseEstimator,
    KerasBaseEstimator,
    create_keras_timeseriesgenerator,
)
from gordo.machine.model.factories import lstm_autoencoder
from gordo.machine.model.base import GordoBase
from gordo.machine.model.register import register_model_builder


logger = logging.getLogger(__name__)


# Generator of model types to each available model kind for that type
# ie. (('KerasAutoEncoder', 'hourglass'), ('KerasAutoEncoder', 'symmetric'), ...)
MODEL_COMBINATIONS = list(
    (model, kind)
    for model in register_model_builder.factories.keys()
    for kind in register_model_builder.factories[model].keys()
)

# Generator of model types to one kind, rather than all combinations of kinds per type
MODEL_SINGLE_KIND = list(
    (model, sorted(register_model_builder.factories[model].keys())[0])
    for model in register_model_builder.factories.keys()
)


@pytest.mark.parametrize("BaseModel", [KerasLSTMBaseEstimator, KerasBaseEstimator])
def test_base_class_models(BaseModel):
    """
    Test that the ABC cannot be instantiated, in that they require some implementation
    """
    with pytest.raises(TypeError):
        BaseModel()


@pytest.mark.parametrize("n_features_out", (2, 3))
@pytest.mark.parametrize("model,kind", MODEL_SINGLE_KIND)
def test_keras_autoencoder_scoring(model, kind, n_features_out):
    """
    Test the KerasAutoEncoder and KerasLSTMAutoEncoder have a working scoring function
    """
    Model = pydoc.locate(f"gordo.machine.model.models.{model}")
    model = Pipeline([("model", Model(kind=kind))])
    X = np.random.random((8, 2))

    # Should be able to deal with y output different than X input features
    y = np.random.random((8, n_features_out))

    with pytest.raises(NotFittedError):
        model.score(X, y)

    model.fit(X, y)
    score = model.score(X, y)
    logger.info(f"Score: {score:.4f}")


@pytest.mark.parametrize("model,kind", MODEL_SINGLE_KIND)
def test_keras_autoencoder_crossval(model, kind):
    """
    Test ability for cross validation
    """
    Model = pydoc.locate(f"gordo.machine.model.models.{model}")
    model = Pipeline([("model", Model(kind=kind))])

    X = np.random.random(size=(15, 2))
    y = X.copy()

    scores = cross_val_score(
        model, X, y, cv=TimeSeriesSplit(n_splits=2, max_train_size=2)
    )
    assert isinstance(scores, np.ndarray)
    logger.info(f"Mean score: {scores.mean():.4f} - Std score: {scores.std():.4f}")


@pytest.mark.parametrize("model,kind", MODEL_COMBINATIONS)
def test_keras_type_config(model, kind):
    """
    Test creating a keras based model from config
    """
    config = {"type": model, "kind": kind}

    # Ensure we can poke the model the same
    model_out = get_model(config)
    assert isinstance(model_out, GordoBase)
    assert isinstance(model_out, BaseWrapper)
    assert isinstance(model_out, pydoc.locate(f"gordo.machine.model.models.{model}"))


@pytest.mark.parametrize("model,kind", MODEL_SINGLE_KIND)
def test_save_load(model, kind):
    config = {"type": model, "kind": kind}

    # Have to call fit, since model production is lazy
    X = np.random.random(size=10).reshape(5, 2)
    y = X.copy()

    # AutoEncoder is fine without a y target
    config["type"] = model
    model_out = get_model(config)
    if model == KerasLSTMForecast:
        assert "forecast_steps" in model_out.get_metadata()

    assert "history" not in model_out.get_metadata()
    model_out.fit(X, y)
    assert "history" in model_out.get_metadata()

    xTest = np.random.random(size=6).reshape(3, 2)
    xHat = model_out.predict(xTest)

    model_out_clone = pickle.loads(pickle.dumps(model_out))

    # Assert parameters are the same.
    assert model_out_clone.get_params() == model_out_clone.get_params()

    # Assert it maintained the state by ensuring predictions are the same
    assert np.allclose(xHat.flatten(), model_out_clone.predict(xTest).flatten())

    assert "history" in model_out.get_metadata()
    assert (
        model_out.get_metadata() == model_out_clone.get_metadata()
    ), "Metadata from model is not same after saving and loading"

    # Assert that epochs list, history dict and params dict in
    # the History object are the same
    assert (
        model_out.model.history.epoch == model_out_clone.model.history.epoch
    ), "Epoch lists differ between original and loaded model history"

    assert (
        model_out.model.history.history == model_out_clone.model.history.history
    ), "History dictionary with losses and accuracies differ between original and loaded model history"

    assert (
        model_out.model.history.params == model_out_clone.model.history.params
    ), "Params dictionaries differ between original and loaded model history"


def test_lookback_window_ae_valueerror_during_fit():
    """
    Assert (for LSTMAutoEncoder) ValueError
    is raised in fit method if lookback_window > number of readings (rows of X)
    """
    lookback_window = 11
    with pytest.raises(ValueError):
        model = KerasLSTMAutoEncoder(
            kind=lstm_autoencoder.lstm_model, lookback_window=lookback_window
        )
        X, y = np.random.rand(10), np.random.rand(10)
        model.fit(X, y)


def test_keras_ae_reshapes_array():
    """
    Asserts KerasLSTMAutoEncoder accepts an array of elements, which it will
    reshape into the matrix of single elements it needs
    """
    model = KerasLSTMAutoEncoder(kind=lstm_autoencoder.lstm_model)
    X, y = np.random.rand(100), np.random.rand(100)
    model.fit(X, y)
    model.predict(X)


def test_keras_forecast_reshapes_array():
    """
    Asserts KerasLSTMForecast accepts an array of elements, which it will
    reshape into the matrix of single elements it needs
    """
    model = KerasLSTMForecast(kind=lstm_autoencoder.lstm_model)
    X, y = np.random.rand(100), np.random.rand(100)
    model.fit(X, y)
    model.predict(X)


def test_lookback_window_ae_valueerror_during_predict():
    """
    Assert (for LSTMAutoEncoder) ValueError
    is raised in fit method if lookback_window > number of readings (rows of X)
    """
    model = KerasLSTMAutoEncoder(kind=lstm_autoencoder.lstm_model, lookback_window=3)
    xTrain, yTrain = np.random.random(size=(4, 2)), np.random.random(size=(4, 2))
    model.fit(xTrain, yTrain)
    with pytest.raises(ValueError):
        model.predict(xTrain[-3:-1, :])


@pytest.mark.parametrize("lookback_window", (5, 6))
def test_lookback_window_forecast_valueerror_during_fit(lookback_window):
    """
    Assert (for LSTMForecast) ValueError is raised
    in fit method if lookback_window >= number of readings (rows of X)
    """
    model = KerasLSTMForecast(
        kind=lstm_autoencoder.lstm_model, lookback_window=lookback_window
    )
    with pytest.raises(ValueError):
        X = np.random.random(size=(5, 2))
        y = X.copy()
        model.fit(X, y)


@pytest.mark.parametrize("lookback_window", (5, 6))
def test_lookback_window_forecast_valueerror_during_predict(lookback_window: int):
    """
    Assert (for LSTMForecast) ValueError is raised
    in fit method if lookback_window >= number of readings (rows of X)
    """
    X = np.random.random(size=(5, 2))
    y = X.copy()
    model = KerasLSTMForecast(
        kind=lstm_autoencoder.lstm_model, lookback_window=lookback_window
    )
    with pytest.raises(ValueError):
        model.fit(X, y)


def test_create_keras_timeseriesgenerator_lb3_loah0_bs2():
    """Check that right output is generated from create_keras_timeseriesgenerator"""
    X = np.array([[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]])
    y = X.copy()
    gen = create_keras_timeseriesgenerator(
        X, y, batch_size=2, lookback_window=3, lookahead=0
    )
    batch_1 = gen[0]
    batch_2 = gen[1]

    batch_1_x = batch_1[0].tolist()
    batch_1_y = batch_1[1].tolist()

    batch_2_x = batch_2[0].tolist()
    batch_2_y = batch_2[1].tolist()

    assert [[[0, 1], [2, 3], [4, 5]], [[2, 3], [4, 5], [6, 7]]] == batch_1_x
    assert [[4, 5], [6, 7]] == batch_1_y

    assert [[[4, 5], [6, 7], [8, 9]]] == batch_2_x
    assert [[8, 9]] == batch_2_y


def test_create_keras_timeseriesgenerator_lb2_loah1_bs2():
    """
    Check right output is generated from create_keras_timeseriesgenerator
    We use lookback_window 2 to get some more interesting batches with lookahead 1
    """
    X = np.array([[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]])
    y = X.copy()
    gen = create_keras_timeseriesgenerator(
        X, y, batch_size=2, lookback_window=2, lookahead=1
    )
    batch_1 = gen[0]
    batch_2 = gen[1]

    batch_1_x = batch_1[0].tolist()
    batch_1_y = batch_1[1].tolist()

    batch_2_x = batch_2[0].tolist()
    batch_2_y = batch_2[1].tolist()

    assert [[[0, 1], [2, 3]], [[2, 3], [4, 5]]] == batch_1_x
    assert [[4, 5], [6, 7]] == batch_1_y

    assert [[[4, 5], [6, 7]]] == batch_2_x
    assert [[8, 9]] == batch_2_y


def test_create_keras_timeseriesgenerator_lb3_loah2_bs2():
    """
    Check right output is generated from create_keras_timeseriesgenerator
    We use lookback_window 2 to get some more interesting batches with lookahead 1
    """
    X = np.array([[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]])
    y = X.copy()
    gen = create_keras_timeseriesgenerator(
        X, y, batch_size=2, lookback_window=2, lookahead=2
    )
    batch_1 = gen[0]
    batch_2 = gen[1]

    batch_1_x = batch_1[0].tolist()
    batch_1_y = batch_1[1].tolist()

    batch_2_x = batch_2[0].tolist()
    batch_2_y = batch_2[1].tolist()

    assert [[[0, 1], [2, 3]], [[2, 3], [4, 5]]] == batch_1_x
    assert [[6, 7], [8, 9]] == batch_1_y

    assert [] == batch_2_x  # No more elements left
    assert [] == batch_2_y


def test_create_keras_timeseriesgenerator_raise_error_on_neg_lookahead():
    """Check create_keras_timeseriesgenerator raises an error on negative lookahead"""
    X = np.array([[0, 1]])
    y = X.copy()
    with pytest.raises(ValueError):
        create_keras_timeseriesgenerator(
            X, y, batch_size=2, lookback_window=2, lookahead=-1
        )


def test_lstmae_predict_output():
    """
    test for KerasLSTMAutoEncoder
      - test dimension of output
      - test that first half of output is testing data
    """
    xTrain, yTrain = np.random.random(size=(5, 3)), np.random.random((5, 3))
    lookback_window = 3
    model = KerasLSTMAutoEncoder(
        kind=lstm_autoencoder.lstm_model, lookback_window=lookback_window
    )
    model = model.fit(xTrain, yTrain)
    xTest = np.random.random(size=(4, 3))
    out = model.predict(xTest)
    assert out.shape == (2, 3)



# -*- coding: utf-8 -*-

import pytest
import numpy as np

from sklearn.ensemble import RandomForestClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from sklearn.decomposition import PCA

from gordo.server import model_io


@pytest.mark.parametrize(
    "model",
    [
        RandomForestClassifier(n_estimators=10),
        Pipeline([("clf", RandomForestClassifier(n_estimators=10))]),
        Pipeline(
            [("mm", MinMaxScaler()), ("clf", RandomForestClassifier(n_estimators=10))]
        ),
        Pipeline([("pca", PCA())]),
    ],
)
def test_model_mixin_get_model_output(model):
    """
    Test ModelMixin which ML Server uses to get model outputs. Basically every model, whether in a Pipeline
    or not, should give back a numpy array; even if the last model doesn't implement predict and just transform.
    """
    X, y = np.random.random((10, 10)), np.random.randint(low=0, high=4, size=10)
    model.fit(X, y)
    out = model_io.get_model_output(model, X)
    assert isinstance(out, np.ndarray)



import re
import logging
from collections import namedtuple
from typing import Union, List, Dict


logger = logging.getLogger(__name__)

SensorTag = namedtuple("SensorTag", ["name", "asset"])

TagPatternToAsset = namedtuple("TagToAsset", ["tag_regexp", "asset_name"])

TAG_TO_ASSET = [
    TagPatternToAsset(re.compile(r"^ninenine.+::.+", re.IGNORECASE), "ninenine"),
    TagPatternToAsset(re.compile(r"^uon_ef.+::.+", re.IGNORECASE), "uon_ef"),
    TagPatternToAsset(re.compile(r"^gfa.", re.IGNORECASE), "1110-gfa"),
    TagPatternToAsset(re.compile(r"^gfb.", re.IGNORECASE), "1111-gfb"),
    TagPatternToAsset(re.compile(r"^gfc.", re.IGNORECASE), "1112-gfc"),
    TagPatternToAsset(re.compile(r"^1125.", re.IGNORECASE), "1125-kvb"),
    TagPatternToAsset(re.compile(r"^tra.", re.IGNORECASE), "1130-troa"),
    TagPatternToAsset(re.compile(r"^asgb.", re.IGNORECASE), "1191-asgb"),
    TagPatternToAsset(re.compile(r"^kri.", re.IGNORECASE), "1175-kri"),
    TagPatternToAsset(re.compile(r"^1138.", re.IGNORECASE), "1138-val"),
    TagPatternToAsset(re.compile(r"^hd.", re.IGNORECASE), "1170-hd"),
    TagPatternToAsset(re.compile(r"^nor.", re.IGNORECASE), "1180-nor"),
    TagPatternToAsset(re.compile(r"^asga.", re.IGNORECASE), "1190-asga"),
    TagPatternToAsset(re.compile(r"^1218.", re.IGNORECASE), "1218-gkr"),
    TagPatternToAsset(re.compile(r"^1219.", re.IGNORECASE), "1219-aha"),
    TagPatternToAsset(re.compile(r"^vis.", re.IGNORECASE), "1230-vis"),
    TagPatternToAsset(re.compile(r"^per.", re.IGNORECASE), "1295-pera"),
    TagPatternToAsset(re.compile(r"^gra.", re.IGNORECASE), "1755-gra"),
    TagPatternToAsset(re.compile(r"^hea.", re.IGNORECASE), "1760-hea"),
    TagPatternToAsset(re.compile(r"^osc.", re.IGNORECASE), "1765-OSC"),
    TagPatternToAsset(re.compile(r"^oss.", re.IGNORECASE), "1766-OSS"),
    TagPatternToAsset(re.compile(r"^ose.", re.IGNORECASE), "1767-OSE"),
    TagPatternToAsset(re.compile(r"^trb.", re.IGNORECASE), "1775-trob"),
    TagPatternToAsset(re.compile(r"^trc.", re.IGNORECASE), "1776-troc"),
    TagPatternToAsset(re.compile(r"^1900.", re.IGNORECASE), "1900-jsv"),
    TagPatternToAsset(re.compile(r"^1901.", re.IGNORECASE), "1901-jsv"),
    TagPatternToAsset(re.compile(r"^1902.", re.IGNORECASE), "1902-jsv"),
    TagPatternToAsset(re.compile(r"^1903.", re.IGNORECASE), "1903-jsv"),
    TagPatternToAsset(re.compile(r"^1904.", re.IGNORECASE), "1904-jsv"),
]


def _asset_from_tag_name(tag_name: str, default_asset: str = None):
    """
    Resolves a tag to the asset it belongs to, if possible.

    If `default_asset` is provided then it will be used as the fallback in case it
    is impossible to resolve tag_name to an asset.

    Parameters
    ----------
    tag_name : str
        Tag to deduce the asset from.
    default_asset : str
        Asset to report if we are not able to resolve tag_name to an assets.

    Returns
    -------
    str
        The asset for provided tag.

    Raises
    ------
    SensorTagNormalizationError
        If we are not able to resolve the tag to an asset and default_asset is not set.
    """
    logger.debug(f"Looking for pattern for tag {tag_name}")

    for pattern in TAG_TO_ASSET:
        if pattern.tag_regexp.match(tag_name):
            logger.debug(
                f"Found pattern {pattern.tag_regexp} in tag {tag_name}, "
                f"returning {pattern.asset_name}"
            )
            return pattern.asset_name
    if default_asset:
        return default_asset
    else:
        raise SensorTagNormalizationError(
            f"Unable to find asset for tag with name {tag_name}"
        )


def _normalize_sensor_tag(
    sensor: Union[Dict, List, str, SensorTag],
    asset: str = None,
    default_asset: str = None,
):
    if isinstance(sensor, Dict):
        return SensorTag(sensor["name"], sensor["asset"])

    elif isinstance(sensor, str):
        if asset is None:
            return SensorTag(
                sensor, _asset_from_tag_name(sensor, default_asset=default_asset)
            )
        else:
            return SensorTag(sensor, asset)

    elif isinstance(sensor, List):
        return SensorTag(sensor[0], sensor[1])

    elif isinstance(sensor, SensorTag):
        return sensor

    raise SensorTagNormalizationError(
        f"Sensor {sensor} with type {type(sensor)} cannot be converted to a valid "
        f"SensorTag"
    )


def normalize_sensor_tags(
    sensors: List[Union[Dict, str, SensorTag]],
    asset: str = None,
    default_asset: str = None,
) -> List[SensorTag]:
    """
    Converts a list of sensors in different formats, into a list of SensorTag elements.

    Notes
    -----
    If you input a list of SensorTag elements, these will just be returned.
    If you input a list of lists or a list of dicts they are expected to contain the
    tag-name and the asset, and no deduction will happen.

    Parameters
    ----------
    sensors : List[Union[Mapping, str, SensorTag]]
            List of sensors
    asset : str
            Optional asset code to put on sensors that don't have it. If this is
            provided we will not attempt to deduce the asset from the tagame.
    default_asset : str
            Optional asset code to put on sensors that dont have it and which we are not
            able to resolve to an asset.

    Returns
    -------
    List[SensorTag]
            List of SensorTags

    """
    logger.debug(
        f"Normalizing list of sensors in some format into SensorTags: {sensors}"
    )
    return [
        _normalize_sensor_tag(sensor_tag_element, asset, default_asset)
        for sensor_tag_element in sensors
    ]


def to_list_of_strings(sensor_tag_list: List[SensorTag]):
    return [sensor_tag.name for sensor_tag in sensor_tag_list]


class SensorTagNormalizationError(ValueError):
    """Error indicating that something went wrong normalizing a sensor tag"""

    pass



import logging
import pandas as pd
import numpy as np

logger = logging.getLogger(__name__)


def apply_buffer(mask: pd.Series, buffer_size: int = 0):
    """
    Take a mask (boolean series) where True indicates keeping a value, and False
    represents removing the value. This will 'expand' those indexes marked as `False`
    to the symmetrical bounds of ``buffer_size``

    Parameters
    ----------
    mask: pandas.core.Series
        Boolean pandas series
    buffer_size: int
        Size to buffer around ``False`` values

    Examples
    --------
    >>> import pandas as pd
    >>> series = pd.Series([True, True, False, True, True])
    >>> apply_buffer(series, buffer_size=1)
    >>> series
    0     True
    1    False
    2    False
    3    False
    4     True
    dtype: bool

    Returns
    -------
    None
    """
    idxs, *_rows = np.where(mask == False)
    for idx in idxs:
        mask.values[
            range(max((0, idx - buffer_size)), min((len(mask), idx + buffer_size + 1)))
        ] = False


def pandas_filter_rows(df, filter_str: str, buffer_size: int = 0):
    """

    Parameters
    ----------
    df: pandas.Dataframe
      Dataframe to filter rows from. Does not modify the parameter
    filter_str: str
      String representing the filter. Can be a boolean combination of conditions,
      where conditions are comparisons of column names and either other columns
      or numeric values. The rows matching the filter are kept.
      Column names with spaces must be quoted with backticks,
      names without spaces could be quoted with backticks or be unquoted.
      Example of legal filters are " `Tag A` > 5 " , " (`Tag B` > 1) | (`Tag C` > 4)"
      '(`Tag D` < 5) ', " (TagB > 5) "
    buffer_size: int
      Area fore and aft of the application of ``fitler_str`` to also mark for removal.

    Returns
    -------
    pandas.Dataframe
        The dataframe containing only rows matching the filter

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> df = pd.DataFrame(list(np.ndindex((3,3))), columns=list('AB'))
    >>> df
       A  B
    0  0  0
    1  0  1
    2  0  2
    3  1  0
    4  1  1
    5  1  2
    6  2  0
    7  2  1
    8  2  2
    >>> pandas_filter_rows(df, "`A`>1")
       A  B
    6  2  0
    7  2  1
    8  2  2
    >>> pandas_filter_rows(df, "`A`> B")
       A  B
    3  1  0
    6  2  0
    7  2  1
    >>> pandas_filter_rows(df, "(`A`>1) | (`B`<1)")
       A  B
    0  0  0
    3  1  0
    6  2  0
    7  2  1
    8  2  2

    """
    pandas_filter = df.eval(filter_str)
    apply_buffer(pandas_filter, buffer_size=buffer_size)
    return df[pandas_filter]



# -*- coding: utf-8 -*-

from typing import List
from unittest.mock import patch

import pandas as pd
import numpy as np

from gordo.client.forwarders import ForwardPredictionsIntoInflux
from gordo.client.utils import influx_client_from_uri
from gordo.machine import Machine
from gordo.machine.dataset import sensor_tag


def get_test_data(columns: List[str]) -> pd.DataFrame:
    """
    Generate timeseries sensor dataframe with given columns
    """
    index = pd.date_range("2019-01-01", "2019-01-02", periods=4)
    df = pd.DataFrame(columns=columns, index=index)

    # Generate some unique values for each key, and insert it into that column
    for i, key in enumerate(columns):
        df[key] = range(i, i + 4)
    return df


def test_influx_forwarder(influxdb, influxdb_uri, sensors, sensors_str):
    """
    Test that the forwarder creates correct points from a
    multi-indexed series
    """
    with patch.object(sensor_tag, "_asset_from_tag_name", return_value="default"):
        machine = Machine.from_config(
            config={
                "name": "some-target-name",
                "dataset": {
                    "tags": sensors_str,
                    "target_tag_list": sensors_str,
                    "train_start_date": "2016-01-01T00:00:00Z",
                    "train_end_date": "2016-01-05T00:00:00Z",
                    "resolution": "10T",
                },
                "model": "sklearn.linear_model.LinearRegression",
            },
            project_name="test-project",
        )

    # Feature outs which match length of tags
    # These should then be re-mapped to the sensor tag names
    input_keys = [("name1", i) for i, _ in enumerate(sensors)]

    # Feature outs which don't match the length of the tags
    # These will be kept at 0..N as field names
    # output_keys = [("name2", f"sensor_{i}") for i in range(len(sensors) * 2)]
    output_keys = [("name2", i) for i in range(len(sensors) * 2)]

    # Assign all keys unique numbers
    df = get_test_data(pd.MultiIndex.from_tuples(input_keys + output_keys))

    # Create the forwarder and forward the 'predictions' to influx.
    forwarder = ForwardPredictionsIntoInflux(destination_influx_uri=influxdb_uri)
    forwarder.forward_predictions(predictions=df, machine=machine)

    # Client to manually verify the points written
    client = influx_client_from_uri(influxdb_uri, dataframe_client=True)

    name1_results = client.query("SELECT * FROM name1")["name1"]

    # Should have column names: 'machine', 'sensor_name', 'sensor_value'
    assert all(
        c in name1_results.columns for c in ["machine", "sensor_name", "sensor_value"]
    )

    # Check that values returned from InfluxDB match what put in for inputs
    for i, tag in enumerate(sensors_str):
        results_mask = name1_results["sensor_name"] == tag
        assert np.allclose(
            df[("name1", i)].values, name1_results[results_mask]["sensor_value"].values
        )

    # Now check the other top level name "name2" is a measurement with the correct points written
    name2_results = client.query("SELECT * FROM name2")["name2"]

    # Should have the same names as tags, since all top levels get stacked into the same resulting columns
    assert all(
        [c in name2_results.columns for c in ["machine", "sensor_name", "sensor_value"]]
    )

    # Check that values returned from InfluxDB match what put in for outputs
    # Note that here the influx sensor names for the output tags are string-cast integers
    for key in output_keys:
        results_mask = name2_results["sensor_name"] == str(key[1])
        assert np.allclose(
            df[key].values, name2_results[results_mask]["sensor_value"].values
        )


def test_influx_send_data(influxdb, influxdb_uri, sensors, sensors_str):
    """
    """
    df = get_test_data(sensors_str)

    # Create the forwarder and forward the sensor data to influx.
    forwarder = ForwardPredictionsIntoInflux(destination_influx_uri=influxdb_uri)
    forwarder.send_sensor_data(df)

    # Client to manually verify the points written
    client = influx_client_from_uri(influxdb_uri, dataframe_client=True)
    resampled_results = client.query("SELECT * FROM resampled")["resampled"]

    # Should have column names: 'sensor_name', 'sensor_value'
    assert all(c in resampled_results.columns for c in ["sensor_name", "sensor_value"])

    # Check that values returned from InfluxDB match what put in for inputs
    for key in sensors_str:
        results_mask = resampled_results["sensor_name"] == key
        assert np.allclose(
            df[key].values, resampled_results[results_mask]["sensor_value"].values
        )



# -*- coding: utf-8 -*-

import io
import logging
import timeit
import typing

from flask import Blueprint, make_response, jsonify, g, request, send_file
from flask_restplus import fields

from gordo import __version__
from gordo.server.rest_api import Api
from gordo.server.views.base import BaseModelView
from gordo.server import utils


logger = logging.getLogger(__name__)

anomaly_blueprint = Blueprint("ioc_anomaly_blueprint", __name__)

api = Api(
    app=anomaly_blueprint,
    title="Gordo API IOC Anomaly Docs",
    version=__version__,
    description="Documentation for the Gordo ML Server",
    default_label="Gordo Endpoints",
)

# POST type declarations
API_MODEL_INPUT_POST = api.model(
    "Prediction - Multiple Samples", {"X": fields.List(fields.List(fields.Float))}
)
API_MODEL_OUTPUT_POST = api.model(
    "Prediction - Output from POST", {"output": fields.List(fields.List(fields.Float))}
)
_tags = {
    fields.String: fields.Float
}  # tags of single prediction record {'tag-name': tag-value}
_single_prediction_record = {
    "start": fields.DateTime,
    "end": fields.DateTime,
    "tags": fields.Nested(_tags),
    "total_abnormality": fields.Float,
}


class AnomalyView(BaseModelView):
    """
    Serve model predictions via POST method.

    Gives back predictions looking something like this
    (depending on anomaly model being served)::

        {
        'data': [
            {
           'end': ['2016-01-01T00:10:00+00:00'],
           'tag-anomaly-scaled': [0.913027075986948,
                                  0.3474043585419292,
                                  0.8986610906818544,
                                  0.11825221990818557],
           'tag-anomaly-unscaled': [10.2335327305725986948,
                                  4.234343958392+3293,
                                  10.379394390232232,
                                  3.32093438982743929],
           'model-output': [0.0005317790200933814,
                            -0.0001525811239844188,
                            0.0008310950361192226,
                            0.0015755111817270517],
           'original-input': [0.9135588550070414,
                              0.3472517774179448,
                              0.8994921857179736,
                              0.11982773108991263],
           'start': ['2016-01-01T00:00:00+00:00'],
           'total-anomaly-unscaled': [1.3326228173185086],
           'total-anomaly-scaled': [0.3020328328002392],
            },
            ...
        ],

     'tags': [{'asset': None, 'name': 'tag-0'},
              {'asset': None, 'name': 'tag-1'},
              {'asset': None, 'name': 'tag-2'},
              {'asset': None, 'name': 'tag-3'}],
     'time-seconds': '0.1937'}
    """

    methods = ["POST"]

    @api.response(200, "Success", API_MODEL_OUTPUT_POST)
    @api.expect(API_MODEL_INPUT_POST, validate=False)
    @api.doc(
        params={
            "X": "Nested list of samples to predict, or single list considered as one sample"
        }
    )
    @utils.model_required
    @utils.extract_X_y
    def post(self):
        start_time = timeit.default_timer()
        return self._create_anomaly_response(start_time)

    def _create_anomaly_response(self, start_time: float = None):
        """
        Use the current ``X`` and ``y`` to create an anomaly specific response
        using the trained ML model's ``.anomaly()`` method.

        Parameters
        ----------
        start_time: Optional[float]
            Start time to use when timing the processing time of the request, will construct a new
            one if not provided.

        Returns
        -------
        flask.Response
            The formatted anomaly representation response object.
        """
        if start_time is None:
            start_time = timeit.default_timer()

        # To use this endpoint, we need a 'y' to calculate the errors.
        if g.y is None:
            message = {
                "message": "Cannot perform anomaly without 'y' to compare against."
            }
            return make_response((jsonify(message), 400))

        # Now create an anomaly dataframe from the base response dataframe
        try:
            anomaly_df = g.model.anomaly(g.X, g.y, frequency=self.frequency)
        except AttributeError:
            msg = {
                "message": f"Model is not an AnomalyDetector, it is of type: {type(g.model)}"
            }
            return make_response(jsonify(msg), 422)  # 422 Unprocessable Entity

        if request.args.get("format") == "parquet":
            return send_file(
                io.BytesIO(utils.dataframe_into_parquet_bytes(anomaly_df)),
                mimetype="application/octet-stream",
            )
        else:
            context: typing.Dict[typing.Any, typing.Any] = dict()
            context["data"] = utils.dataframe_to_dict(anomaly_df)
            context["time-seconds"] = f"{timeit.default_timer() - start_time:.4f}"
            return make_response(jsonify(context), context.pop("status-code", 200))


api.add_resource(
    AnomalyView, "/gordo/v0/<gordo_project>/<gordo_name>/anomaly/prediction"
)



# -*- coding: utf-8 -*-

import math
from typing import Tuple


def hourglass_calc_dims(
    compression_factor: float, encoding_layers: int, n_features: int
) -> Tuple[int, ...]:
    """
    Calculate the layer dimensions given the number of layers, compression, and features

    Parameters
    ----------
    compression_factor: float
        How small the smallest layer is as a ratio of n_features
        (smallest layer is rounded up to nearest integer). Must satisfy
        0 <= compression_factor <= 1.
    encoding_layers: int
        Number of layers from the input layer (exclusive) to the
        narrowest layer (inclusive). Must be > 0. The total nr of layers
        including input and output layer will be 2*encoding_layers + 1.
    n_features_out: Optional[int]
        Number of features the model will output, default to ``n_features``.

    Returns
    -------
    dims: Tuple[int,...]
         Number of neurons per layers for the encoder, reversed for the decoder.
         Must have len > 0
    """
    if not (1 >= compression_factor >= 0):
        raise ValueError("compression_factor must be 0 <= compression_factor <= 1")
    if encoding_layers < 1:
        raise ValueError("encoding_layers must be >= 1")
    smallest_layer = max(min(math.ceil(compression_factor * n_features), n_features), 1)
    diff = n_features - smallest_layer
    average_slope = diff / encoding_layers
    dims = tuple(
        round(n_features - (i * average_slope)) for i in range(1, encoding_layers + 1)
    )
    return dims


def check_dim_func_len(prefix: str, dim: Tuple[int, ...], func: Tuple[str, ...]):
    """
    Check that the number of layer dimensions and layer functions are equal

    Parameters
    ----------
    prefix: str
        Parameter name prefix for error message generation (Options: "encoding" or "decoding").
    dim: tuple of int
        Tuple of numbers with the number of neurons in the encoding or decoding part.
    func: Tuple[str,...]
        Tuple of numbers with the number of neurons in the decoding part.
    """
    if len(dim) != len(func):
        raise ValueError(
            f"The length (i.e. the number of network layers) of {prefix}_dim "
            f"({len(dim)}) and {prefix}_func ({len(func)}) must be equal. If only "
            f"{prefix}_dim or {prefix}_func was passed, ensure that its length matches "
            f"that of the {prefix} parameter not passed."
        )
